/***************************************************************************
                          cdownloadmanager.cpp  -  description
                             -------------------
    begin                : Don Mai 16 2002
    copyright            : (C) 2002-2003 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>

#include <dclib/dcos.h>
#include <dclib/dcobject.h>
#include <dclib/cconfig.h>
#include <dclib/cdir.h>
#include <dclib/cservermanager.h>
#include <dclib/cmd5.h>
#include <dclib/clisten.h>
#include <dclib/cdownloadqueue.h>
#include <dclib/clogfile.h>
#include <dclib/cmanager.h>
#include <dclib/filecopy.h>

#ifndef WIN32
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#endif

#define CHUNK_SIZE	(1024*1024)
#define TEST_CHUNK_SIZE	(10*1024)

#include "cdownloadmanager.h"

/** */
CDownloadManager::CDownloadManager()
{
	lTransferID    = time(0);
	m_eShutdownState = essNONE;

	tDownloadQueueTimeout  = time(0);
	tUpdateTransferTimeout = time(0);

	m_pTransferList      = new CThreadStringList();
	m_pTransferBanList   = new CThreadStringList();
	m_pExtraUserSlotList = new CThreadList<CExtraUserSlot>;
	m_pTransferWaitList  = new CThreadList<DCTransferWait>;

	m_pListen = new CListen();
	m_pDownloadQueue = new CDownloadQueue();

	m_pCallback       = 0;

	SetInstance(this);
}

/** */
CDownloadManager::~CDownloadManager()
{
	DLM_StopListen();

	CManager::Instance()->Remove( m_pCallback );
	if ( m_pCallback )
	{
		delete m_pCallback;
		m_pCallback = 0;
	}

	if (m_pExtraUserSlotList)
	{
		delete m_pExtraUserSlotList;
		m_pExtraUserSlotList = 0;
	}

	if(m_pTransferList)
	{
		delete m_pTransferList;
		m_pTransferList = 0;
	}

	if(m_pTransferWaitList)
	{
		delete m_pTransferWaitList;
		m_pTransferWaitList = 0;
	}

	if(m_pTransferBanList)
	{
		delete m_pTransferBanList;
		m_pTransferBanList = 0;
	}

	if ( m_pListen )
	{
		delete m_pListen;
		m_pListen = 0;
	}

	if ( m_pDownloadQueue )
	{
		delete m_pDownloadQueue;
		m_pDownloadQueue = 0;
	}
}

/** load the queue */
int CDownloadManager::DLM_LoadQueue()
{
	int err;
	CStringList * StringList = 0;

	m_pDownloadQueue->pQueue->Lock();
	m_pDownloadQueue->pQueue->Clear();
	m_pDownloadQueue->pChunkList->Lock();
	m_pDownloadQueue->pChunkList->Clear();

	err = CConfig::Instance()->LoadDCTra( m_pDownloadQueue->pQueue, m_pDownloadQueue->pChunkList );

	// send all files over the callback function
	if ( err == 0 )
	{
		while( m_pDownloadQueue->pQueue->Next( (CObject *&) StringList) )
		{
			DCTransferQueueObject * TransferObject = 0;

			while( StringList->Next( (CObject *&) TransferObject) )
			{
				DCTransferFileObject * TransferFileObject = 0;

				while( TransferObject->pTransferFileList.Next( (CObject *&)TransferFileObject) )
				{
					SendFileInfo( TransferObject, TransferFileObject );
				}
			}
		}
	}

	m_pCallback = new CCallback<CDownloadManager>( this, &CDownloadManager::Callback );
	CManager::Instance()->Add( m_pCallback );

	m_pDownloadQueue->pChunkList->UnLock();
	m_pDownloadQueue->pQueue->UnLock();

	return err;
}

/** save the queue  */
int CDownloadManager::DLM_SaveQueue()
{
	int err;

	m_pDownloadQueue->pQueue->Lock();
	m_pDownloadQueue->pChunkList->Lock();

	err = CConfig::Instance()->SaveDCTra( m_pDownloadQueue->pQueue, m_pDownloadQueue->pChunkList );

	m_pDownloadQueue->pChunkList->UnLock();
	m_pDownloadQueue->pQueue->UnLock();

	return err;
}

/** disconnect all transfers for a clean shutdown */
void CDownloadManager::DLM_Shutdown()
{
	CObject *obj;
	CTransferObject * TransferObject;

	// stop listen
	DLM_StopListen();

	// set shutdown state
	m_eShutdownState = essSHUTDOWN;

	SendLogInfo("Shutdown download manager ...\n");

	m_pTransferList->Lock();

	printf("Running Transfers: %ld\n",m_pTransferList->Count());

	obj = 0;
	while( m_pTransferList->Next( (CObject*&)obj ) )
	{
		TransferObject = (CTransferObject*)obj;
		TransferObject->m_pTransfer->Disconnect(TRUE);
	}

	m_pTransferList->UnLock();
}

/** */
bool CDownloadManager::DLM_InitListen( CString & s )
{
	bool res = FALSE;

	if ( m_pListen->StartListen( CConfig::Instance()->GetTCPListenPort() ) == 0 )
	{
		m_pListen->SetCallBackFunction( new CCallback<CDownloadManager>( this, &CDownloadManager::DM_ListenCallBack ) );

		res = TRUE;
	}
	else
	{
		s = m_pListen->GetSocketError();
	}

	return res;
}

/** */
void CDownloadManager::DLM_StopListen()
{
	m_pListen->SetCallBackFunction(0);
	m_pListen->StopListen();
}

/** add a extra slot to the user */
void CDownloadManager::DLM_AddUserSlot( CString nick, CString hubname, int slot, bool permanent )
{
	CExtraUserSlot * ExtraUserSlot;

	m_pExtraUserSlotList->Lock();

	ExtraUserSlot = 0;

	while( (ExtraUserSlot=m_pExtraUserSlotList->Next(ExtraUserSlot)) != 0 )
	{
		if ( (ExtraUserSlot->sNick == nick) && (ExtraUserSlot->sHubName == hubname) )
		{
			if ( (slot == 0) && (!permanent) )
			{
				ExtraUserSlot->iSlots     = slot;
				ExtraUserSlot->bPermanent = FALSE;
			}
			else
			{
				if (ExtraUserSlot->bPermanent)
					ExtraUserSlot->iSlots = 0;
				ExtraUserSlot->iSlots     += slot;
				ExtraUserSlot->bPermanent  = permanent;
			}
			break;
		}
	}

	if ( ExtraUserSlot == 0 )
	{
		ExtraUserSlot = new CExtraUserSlot();
		ExtraUserSlot->sNick      = nick;
		ExtraUserSlot->sHubName   = hubname;
		ExtraUserSlot->iSlots     = slot;
		ExtraUserSlot->bPermanent = permanent;

		m_pExtraUserSlotList->Add(ExtraUserSlot);
	}

	SendSlotInfo(ExtraUserSlot);

	if ( (ExtraUserSlot->iSlots == 0) && (!ExtraUserSlot->bPermanent) )
	{
		m_pExtraUserSlotList->Del(ExtraUserSlot);
	}

	m_pExtraUserSlotList->UnLock();
}

/** */
bool CDownloadManager::DLM_TransferConnect( CString nick, CString hubname )
{
	bool res = FALSE;

	m_pDownloadQueue->pQueue->Lock();

	DCTransferQueueObject * TransferObject;

	if ( (TransferObject = m_pDownloadQueue->GetUserTransferObject( nick, hubname )) != 0 )
	{
		TransferObject->tTimeout = 0;
		res = TRUE;
	}

	m_pDownloadQueue->pQueue->UnLock();

	return res;
}

/** */
bool CDownloadManager::DLM_TransferClose( CString transferid )
{
	CObject *obj;
	CTransferObject * TransferObject;
	bool res = FALSE;

	m_pTransferList->Lock();

	if ( m_pTransferList->Get( transferid, (CObject*&)obj) == 0 )		
	{
		TransferObject = (CTransferObject*)obj;
		TransferObject->m_pTransfer->Disconnect(TRUE);
		res = TRUE;
	}

	m_pTransferList->UnLock();

	return res;
}

/** */
bool CDownloadManager::DLM_TransferSetRate( CString transferid, ulonglong rate )
{
	CObject * obj;
	CTransferObject * TransferObject;
	bool res = FALSE;

	m_pTransferList->Lock();

	if ( m_pTransferList->Get( transferid, (CObject*&)obj ) == 0 )
	{
		TransferObject = (CTransferObject*)obj;
		TransferObject->m_pTransfer->SetRate(rate);
		res = TRUE;
	}

	m_pTransferList->UnLock();

	return res;
}

/** */
bool CDownloadManager::DLM_TransferGetRate( CString transferid, ulonglong & rate )
{
	CObject * obj;
	CTransferObject * TransferObject;
	bool res = FALSE;

	m_pTransferList->Lock();

	if ( m_pTransferList->Get( transferid, (CObject*&)obj) == 0 )
	{
		TransferObject = (CTransferObject*)obj;
		rate = TransferObject->m_pTransfer->GetRate();
		res = TRUE;
	}

	m_pTransferList->UnLock();

	return res;
}

/** */
eDirection CDownloadManager::DLM_TransferDirection( CString transferid )
{
	CObject * obj;
	CTransferObject * TransferObject;
	eDirection direction=edNONE;

	m_pTransferList->Lock();

	if ( m_pTransferList->Get( transferid, (CObject*&)obj ) == 0 )
	{
		TransferObject = (CTransferObject*)obj;
		direction = TransferObject->m_pTransfer->GetSrcDirection();
	}

	m_pTransferList->UnLock();

	return direction;
}

/** */
void CDownloadManager::DLM_QueueAdd( CString nick, CString hubname, CString hubhost,
				CString remotename, CString localname,
				CString localpath, CString localrootpath,
				eltMedium medium, ulonglong size,
				ulonglong startposition,
				bool multi )
{
	CDir dir;
	CString sfile;
	DCTransferQueueObject * TransferObject;
	DCTransferFileObject * TransferFileObject = 0;
	DCHubObject * HubObject;

	printf("add queue entry: %s %s %s %s %d %s %s %s %llu %llu\n",
		nick.Data(),
		remotename.Data(),
		hubname.Data(),
		hubhost.Data(),
		medium,
		localname.Data(),
		localpath.Data(),
		localrootpath.Data(),
		startposition,
		size );

	m_pDownloadQueue->pQueue->Lock();

	CStringList * StringList = m_pDownloadQueue->GetUserHubList( nick );

	if ( (TransferObject = m_pDownloadQueue->GetUserTransferObject( nick, hubname )) == 0 )
	{
		TransferObject = new DCTransferQueueObject();

		TransferObject->sNick    = nick;
		TransferObject->sHubHost = hubhost;
		TransferObject->sHubName = hubname;
		TransferObject->eState   = etwsIDLE;
		TransferObject->iConnections = 0;
		TransferObject->tTimeout = 0;

		HubObject = new DCHubObject();

		HubObject->m_sHubName = hubname;
		HubObject->m_sHubHost = hubhost;
		HubObject->m_bActive  = TRUE;

		TransferObject->pHubList.Add(HubObject);

		if ( StringList == 0 )
		{
			StringList = new CStringList();
			m_pDownloadQueue->pQueue->Add( nick, StringList );
		}

		StringList->Add( hubname, TransferObject );

		printf("transfer add\n");
	}
	else
	{
		TransferFileObject = m_pDownloadQueue->GetUserFileObject( nick, hubname, remotename );
	}

	// file not found for the nick/hub
	if ( TransferFileObject == 0 )
	{
		TransferFileObject = new DCTransferFileObject();

		TransferFileObject->m_eState      = etfsNONE;
		TransferFileObject->m_nSize       = size;
		TransferFileObject->m_bMulti      = multi;
		TransferFileObject->m_eMedium     = medium;
		TransferFileObject->m_sRemoteFile = remotename;

		if ( remotename == CString(DC_USER_FILELIST) )
		{
			TransferFileObject->m_nPriority = 0;
		}
		else
		{
			TransferFileObject->m_nPriority = 2;
		}
			
		if ( localrootpath == "" )
		{
			sfile = CConfig::Instance()->GetDownloadFolder();
		}
		else
		{
			sfile = localrootpath;
		}

		sfile = sfile + "/" + localpath + "/" + localname;
		sfile = dir.SimplePath(sfile);

		TransferFileObject->m_sLocalFile     = sfile;
		TransferFileObject->m_sLocalPath     = localpath;
		TransferFileObject->m_sLocalFileName = localname;

		TransferObject->pTransferFileList.Add( remotename, TransferFileObject );

		printf("file add '%s' %ld\n",sfile.Data(),TransferObject->pTransferFileList.Count());

		m_pDownloadQueue->pChunkList->Lock();

		DCFileChunkObject * FileChunkObject;

		if ( remotename == CString(DC_USER_FILELIST) )
		{
			printf("no chunk for userlists\n");
		}
		else if ( m_pDownloadQueue->pChunkList->Get( sfile, (CObject *&)FileChunkObject ) != 0 )
		{
			FileChunkObject = new DCFileChunkObject();

			FileChunkObject->sLocalFile      = sfile;
			FileChunkObject->stHash          = "";
			FileChunkObject->bMulti          = multi;
			FileChunkObject->lSize           = size;
			FileChunkObject->lSizeDone       = startposition;
			FileChunkObject->iReferenceCount = 1;

			// create the first unfinished chunk
			DCChunkObject * ChunkObject = new DCChunkObject();
			ChunkObject->lStart = startposition;
			ChunkObject->lEnd   = size;
			FileChunkObject->pChunks.Add(ChunkObject);

			m_pDownloadQueue->pChunkList->Add( sfile, FileChunkObject );

			printf("add file chunk object\n");
		}
		else
		{
			FileChunkObject->iReferenceCount++;
			printf("file chunk object found\n");
		}

		m_pDownloadQueue->pChunkList->UnLock();
	}
	else
	{
		printf("file found ...\n");
	}

	SendFileInfo( TransferObject, TransferFileObject );

	m_pDownloadQueue->pQueue->UnLock();
}

/**	0: file not in the queue
	1: file allready in the queue by same nick und hub
	2: local file with same size allready in the queue (multi download)
	3: local file with same size allready in the queue without multi (no multi download)
	4: local file with not the same size allready in the queue (multi download error)
*/
int CDownloadManager::DLM_QueueCheck( CString nick, CString hubname, CString /*hubhost*/,
				CString remotename, CString localname,
				CString localpath, CString localrootpath,
				eltMedium, ulonglong size )
{
	m_pDownloadQueue->pQueue->Lock();

	CDir dir;
	CString sfile;
	int res = 0;
	DCTransferQueueObject * TransferObject;
	DCTransferFileObject * TransferFileObject = 0;

	if ( (TransferObject = m_pDownloadQueue->GetUserTransferObject( nick, hubname )) != 0 )
	{
		if ( (TransferFileObject = m_pDownloadQueue->GetUserFileObject( nick, hubname, remotename )) != 0 )
		{
			res = 1;
		}
	}

	// check if a local file allready exist
	if ( (res == 0) && (remotename != CString(DC_USER_FILELIST)) )
	{
		if ( localrootpath == "" )
		{
			sfile = CConfig::Instance()->GetDownloadFolder();
		}
		else
		{
			sfile = localrootpath;
		}

		sfile = sfile + "/" + localpath + "/" + localname;
		sfile = dir.SimplePath(sfile);

		DCFileChunkObject * FileChunkObject;

		m_pDownloadQueue->pChunkList->Lock();

		if ( m_pDownloadQueue->pChunkList->Get( sfile, (CObject *&)FileChunkObject ) == 0 )
		{
			if ( FileChunkObject->lSize == size )
			{
				if ( FileChunkObject->bMulti == TRUE )
				{
					res = 2;
				}
				else
				{
					res = 3;
				}
			}
			else
			{
				res = 4;
			}
		}

		m_pDownloadQueue->pChunkList->UnLock();
	}

	m_pDownloadQueue->pQueue->UnLock();

	return res;
}

/** change the src-nick/hubname to dst-nick/hubname in the queue */
bool CDownloadManager::DLM_QueueEdit( CString srcnick, CString srchubname, CString dstnick, CString dsthubname, CString dsthubhost )
{
	bool res;
	DCTransferQueueObject * TransferObject;
	DCTransferFileObject * TransferFileObject;

	m_pDownloadQueue->pQueue->Lock();

	res = FALSE;

	// check if transfer run
	if ( (TransferObject = m_pDownloadQueue->GetUserTransferObject( srcnick, srchubname )) != 0 )
	{
		if ( (TransferObject->eState != etwsWAIT) &&
		     (TransferObject->eState != etwsRUN) )
		{
			// only rename if the new not exists
			if ( m_pDownloadQueue->GetUserTransferObject( dstnick, dsthubname ) == 0 )
			{
				// send remove all files
				TransferFileObject = 0;

				while( TransferObject->pTransferFileList.Next( (CObject *&)TransferFileObject) )
				{
					SendFileInfo( TransferObject, TransferFileObject, TRUE );
				}

				// rename old entry
				m_pDownloadQueue->RenameNick( srcnick, dstnick, srchubname, dsthubname );

				// get hubhost from new entry
				if ( (TransferObject = m_pDownloadQueue->GetUserTransferObject( dstnick, dsthubname )) != 0 )
				{
					TransferObject->sHubHost = dsthubhost;

					// send new files
					TransferFileObject = 0;

					while( TransferObject->pTransferFileList.Next( (CObject *&)TransferFileObject) )
					{
						SendFileInfo( TransferObject, TransferFileObject );
					}

					res = TRUE;
				}
			}
			else
			{
				// update hubhost
				if ( TransferObject->sHubHost != dsthubhost )
				{
					TransferObject->sHubHost = dsthubhost;
					SendFileInfo( TransferObject );
				}
				
			}
		}
	}

	m_pDownloadQueue->pQueue->UnLock();

	return res;
}

/** */
bool CDownloadManager::DLM_QueueRemove( CString nick, CString hubname, CString remotefile )
{
	bool res;
	
	m_pDownloadQueue->pQueue->Lock();

	res = RemoveQueueFile( nick, hubname, remotefile );

	m_pDownloadQueue->pQueue->UnLock();
	
	return res;
}

/** */
bool CDownloadManager::DLM_QueueRemove( CString localfile )
{
	bool res;

	m_pDownloadQueue->pQueue->Lock();

	res = RemoveQueueFile(localfile);

	m_pDownloadQueue->pQueue->UnLock();

	return res;
}

/** */
bool CDownloadManager::DLM_QueueSetFilePriority( CString nick, CString hubname, CString remotefile, int priority )
{
	m_pDownloadQueue->pQueue->Lock();

	bool res = FALSE;
	DCTransferQueueObject * TransferObject = 0;
	DCTransferFileObject * TransferFileObject = 0;

	if ( (TransferObject = m_pDownloadQueue->GetUserTransferObject( nick, hubname )) != 0 )
	{
		if ( (remotefile != "") && (priority <= MAX_FILE_PRIORITY) )
		{
			if ( (TransferFileObject = m_pDownloadQueue->GetUserFileObject( nick, hubname, remotefile )) != 0 )
			{
				// we don't need check if transfer is running
				TransferFileObject->m_nPriority = priority;

				SendFileInfo( TransferObject, TransferFileObject );

				res = TRUE;
			}
		}
	}

	m_pDownloadQueue->pQueue->UnLock();
	
	return res;
}

/** */
bool CDownloadManager::DLM_QueueGetFileInfo( CString nick, CString hubname, CString /*hubhost*/, CString remotefile, CUserFileInfo * UserFileInfo )
{
	bool res = FALSE;

	if (!UserFileInfo)
	{
		return res;
	}

	m_pDownloadQueue->pQueue->Lock();

	DCTransferQueueObject * TransferObject;
	DCTransferFileObject * TransferFileObject;

	if ( (TransferObject = m_pDownloadQueue->GetUserTransferObject( nick, hubname )) != 0 )
	{
		UserFileInfo->eWaitState    = TransferObject->eState;
		UserFileInfo->sUserFileList = TransferObject->sUserFileList;

		if ( remotefile != "" )
		{
			if ( TransferObject->pTransferFileList.Get( remotefile, (CObject *&) TransferFileObject ) == 0 )
			{
				UserFileInfo->eFileState = TransferFileObject->m_eState;
				UserFileInfo->sLocalFile = TransferFileObject->m_sLocalFile;
				UserFileInfo->bMulti     = TransferFileObject->m_bMulti;
				res = TRUE;
			}
		}
		else
		{
			res = TRUE;
		}
	}

	m_pDownloadQueue->pQueue->UnLock();

	return res;
}

/** */
DCFileChunkObject * CDownloadManager::DLM_QueueGetFileChunk( CString file )
{
	m_pDownloadQueue->pQueue->Lock();

	DCFileChunkObject * FileChunkObject1, * FileChunkObject2=0;

	if ( (FileChunkObject1 = m_pDownloadQueue->GetFileChunkObject(file)) != 0 )
	{
		FileChunkObject2 = new DCFileChunkObject(FileChunkObject1);
	}

	m_pDownloadQueue->pQueue->UnLock();

	return FileChunkObject2;
}

/** */
bool CDownloadManager::DLM_QueueUpdateHub( CString nick, CString hubname )
{
	m_pDownloadQueue->pQueue->Lock();

	bool res = FALSE;
	DCConfigHubItem hubitem;
	DCTransferQueueObject * TransferObject;

	if ( (TransferObject = m_pDownloadQueue->GetUserTransferObject( nick, hubname )) != 0 )
	{
		if ( (res = CConfig::Instance()->GetPublicHub( hubname, &hubitem )) == TRUE )
			TransferObject->sHubHost = hubitem.sHost;
		else if ( (res = CConfig::Instance()->GetBookmarkHub( hubname, &hubitem )) == TRUE )
			TransferObject->sHubHost = hubitem.sHost;

		if ( res == TRUE )
			SendFileInfo( TransferObject );
	}

	m_pDownloadQueue->pQueue->UnLock();

	return res;
}

/** */
void CDownloadManager::DLM_QueueGetHub( CString nick, CString hubname, CList<DCHubObject> * list )
{
	DCTransferQueueObject * TransferObject;
	DCHubObject * HubObject1;

	if (!list)
	{
		return;
	}

	list->Clear();

	m_pDownloadQueue->pQueue->Lock();

	if ( (TransferObject = m_pDownloadQueue->GetUserTransferObject( nick, hubname )) != 0 )
	{
		HubObject1=0;

		while( (HubObject1=TransferObject->pHubList.Next(HubObject1)) != 0 )
			list->Add( new DCHubObject(HubObject1) );
	}

	m_pDownloadQueue->pQueue->UnLock();

	return;
}



/** */
bool CDownloadManager::DLM_AddTransferRequest( CString nick, CString userhost, CString hubname, CString hubhost )
{
	bool res;
	DCTransferWait * TransferWait;

	m_pTransferWaitList->Lock();

	res = FALSE;
	TransferWait = 0;

	// check if the user allready in the waitlist
	while ( (TransferWait = m_pTransferWaitList->Next(TransferWait)) != 0 )
	{
		if ( TransferWait->sHubName == hubname )
		{
			if ( ((TransferWait->sNick == nick) && (TransferWait->sNick != "")) ||
			     ((TransferWait->sUserHost == userhost) && (TransferWait->sUserHost != "")))
			{
				// TODO: count in requests from a client
				// reset timeout
				TransferWait->tTimeout = time(0);

				break;
			}
		}
	}

	if ( TransferWait == 0 )
	{
		// user not found, add it
		TransferWait = new DCTransferWait();
		TransferWait->sNick     = nick;
		TransferWait->sUserHost = userhost;
		TransferWait->sHubName  = hubname;
		TransferWait->sHubHost  = hubhost;
		TransferWait->tTimeout  = time(0);

		m_pTransferWaitList->Add(TransferWait);

		printf("Add wait transfer for: '%s' '%s' '%s'\n",nick.Data(),hubname.Data(),hubhost.Data());
	}

	res = TRUE;

	m_pTransferWaitList->UnLock();

	return res;
}

/** */
void CDownloadManager::DLM_AddTransferRequest( CString host, int port, CString hubname, CString hubhost )
{
	// check private address space
	if ( !((CConfig::Instance()->GetCheckPrivateAddressSpace() == TRUE) &&
	     (CSocket::IsPrivateAddressSpace(host.Data()) == TRUE )) )
	{
		// TODO: check if client allready in the transfer wait list
		// create a new transfer
		CTransferObject * TransferObject = new CTransferObject();
		TransferObject->m_pTransfer = new CTransfer();

		TransferObject->m_pTransfer->SetTransferID( GetNewTransferID() );
		TransferObject->m_pTransfer->SetNick( CConfig::Instance()->GetNick() );

		TransferObject->m_pTransfer->SetHubName(hubname);
		TransferObject->m_pTransfer->SetHubHost(hubhost);

		TransferObject->m_pTransfer->SetHost( host, port );
		TransferObject->m_pTransfer->SetRate( CConfig::Instance()->GetMaxUploadRate() );

		// add this transfer to the wait list
		DLM_AddTransferRequest( "", TransferObject->m_pTransfer->GetHost(), hubname, hubhost );

		m_pTransferList->Lock();

		m_pTransferList->Add( TransferObject->m_pTransfer->GetTransferID(), TransferObject );
		TransferObject->m_pTransfer->SetCallBackFunction( new CCallback<CDownloadManager>( this, &CDownloadManager::DM_TransferCallBack ) );
		// connect to the client
		TransferObject->m_pTransfer->Connect();

		m_pTransferList->UnLock();
	}
	else
	{
		// send warning
		SendLogInfo( "Warning: Detect private address space: " + host + ":" + CString().setNum(port) + " at hub '" + hubname + "' ("+hubhost+")" );
	}
}

/** */
bool CDownloadManager::DLM_GetDownloadManagerInfo( CDownloadManagerInfo * pinfo )
{
	*pinfo = DownloadManagerInfo;

	return TRUE;
}

/** */
void CDownloadManager::SendFileInfo( DCTransferQueueObject * TransferObject, DCTransferFileObject * TransferFileObject, bool bRemoveFile )
{
	LogThread.Lock();

	CMessageDMFileObject * fo = new CMessageDMFileObject();

	fo->sNick       = TransferObject->sNick;
	fo->sHubName    = TransferObject->sHubName;
	fo->sHubHost    = TransferObject->sHubHost;
	fo->eWaitState  = TransferObject->eState;
	fo->tTimeout    = TransferObject->tTimeout;
	fo->bRemoveFile = bRemoveFile;
	fo->iConnections = TransferObject->iConnections;

	if ( TransferFileObject )
	{
		fo->sRemoteFile = TransferFileObject->m_sRemoteFile;
		fo->sLocalFile  = TransferFileObject->m_sLocalFile;
		fo->lSize       = TransferFileObject->m_nSize;
		fo->eFileState  = TransferFileObject->m_eState;
		fo->bMulti      = TransferFileObject->m_bMulti;
		fo->m_nPriority = TransferFileObject->m_nPriority;
	}

	fo->m_eType = DC_MESSAGE_FILE_OBJECT;

	if ( DC_DownloadManagerCallBack(fo) == -1 )
		delete fo;

	LogThread.UnLock();
}

/** */
void CDownloadManager::SendTransferInfo( CTransfer * Transfer, bool remove )
{
	LogThread.Lock();

	DCFileChunkObject * FileChunkObject;
	CMessageDMTransferObject * to = new CMessageDMTransferObject();

	to->sTransferID      = Transfer->GetTransferID();
	to->sNick            = Transfer->GetDstNick();
	to->sHost            = Transfer->GetHost();
	to->sHubName         = Transfer->GetHubName();
	to->eState           = Transfer->GetMode();
	to->sRemoteFile      = Transfer->GetDstFilename();
	to->sRemoteFileName  = "";
	to->lSize            = Transfer->GetLength();
	to->lStartPosition   = Transfer->GetStartPosition();
	to->lCurrentPosition = Transfer->GetCurrentPosition();
	to->lEndPosition     = Transfer->GetEndPosition();

	to->lRate            = Transfer->GetTransferrate();
	to->bRemoveTransfer  = remove;

	// get size done from the chunk list
	if ( Transfer->GetSrcDirection() == edUPLOAD )
	{
		to->lSizeDone = Transfer->GetCurrentPosition();
	}
	else
	{
		m_pDownloadQueue->pChunkList->Lock();

		if ( Transfer->GetMedium() == eltBUFFER )
		{
			to->lSizeDone = Transfer->GetCurrentPosition();
		}
		else if ( (FileChunkObject = m_pDownloadQueue->GetFileChunkObject(Transfer->GetSrcFilename())) != 0 )
		{
			to->lSizeDone = FileChunkObject->lSizeDone+to->lCurrentPosition-to->lStartPosition;
		}
		else
		{
			to->lSizeDone = to->lSize;
		}

		m_pDownloadQueue->pChunkList->UnLock();	
	}

	to->m_eType = DC_MESSAGE_TRANSFER_OBJECT;

	if ( DC_DownloadManagerCallBack(to) == -1 )
		delete to;

	LogThread.UnLock();
}

/** */
void CDownloadManager::SendSlotInfo( CExtraUserSlot * Object )
{
	LogThread.Lock();

	CMessageDMSlotObject * so = new CMessageDMSlotObject();

	so->sNick      = Object->sNick;
	so->sHubName   = Object->sHubName;
	so->iSlots     = Object->iSlots;
	so->bPermanent = Object->bPermanent;

	so->m_eType = DC_MESSAGE_SLOT_OBJECT;

	if ( DC_DownloadManagerCallBack(so) == -1 )
		delete so;

	LogThread.UnLock();
}

/** */
void CDownloadManager::SendLogInfo( CString message, CTransfer * Transfer )
{
	LogThread.Lock();

	CString s = "";

	CMessageLog * log = new CMessageLog();

	if ( Transfer != 0 )
	{
		s = "[";
		if ( Transfer->GetDstNick() == "" )
			s += "???";
		else
			s += Transfer->GetDstNick();
		s += "] ";
	}
	
	s += message;
	
	log->sMessage = s;
	log->m_eType  = DC_MESSAGE_LOG;

	if ( DC_DownloadManagerCallBack(log) == -1 )
		delete log;

	LogThread.UnLock();
}

/** */
void CDownloadManager::SendDownloadManagerInfo( CDownloadManagerInfo * dminfo )
{
	LogThread.Lock();

	CDownloadManagerInfo * dmi = new CDownloadManagerInfo;

	*dmi = *dminfo;

	if ( DC_DownloadManagerCallBack(dmi) == -1 )
		delete dmi;

	LogThread.UnLock();
}

/** */
void CDownloadManager::SendTrafficInfo()
{
	LogThread.Lock();

	DCMessageTraffic * to = new DCMessageTraffic();

	to->m_nRx        = CSocket::m_Traffic.GetTraffic(ettRX);
	to->m_nTx        = CSocket::m_Traffic.GetTraffic(ettTX);
	to->m_nDataRx    = CSocket::m_Traffic.GetTraffic(ettDATARX);
	to->m_nDataTx    = CSocket::m_Traffic.GetTraffic(ettDATATX);
	to->m_nControlRx = CSocket::m_Traffic.GetTraffic(ettCONTROLRX);
	to->m_nControlTx = CSocket::m_Traffic.GetTraffic(ettCONTROLTX);

	if ( DC_DownloadManagerCallBack(to) == -1 )
		delete to;

	LogThread.UnLock();
}

/** */
void CDownloadManager::SendFileManagerInfo( CFileManagerInfo * info )
{
	LogThread.Lock();


	CFileManagerInfo * fmi = new CFileManagerInfo();

	*fmi = *info;

	if ( DC_DownloadManagerCallBack(fmi) == -1 )
		delete fmi;

	LogThread.UnLock();
}

/** */
bool CDownloadManager::CheckUserSlot( CString nick, CString )
{
	bool res;
	CExtraUserSlot * ExtraUserSlot;

	m_pExtraUserSlotList->Lock();

	res = FALSE;
	ExtraUserSlot = 0;

	while( (ExtraUserSlot=m_pExtraUserSlotList->Next(ExtraUserSlot)) != 0 )
	{
		if ( ExtraUserSlot->sNick == nick )
		{
			if ( ExtraUserSlot->iSlots > 0 )
			{
				ExtraUserSlot->iSlots--;

				SendSlotInfo(ExtraUserSlot);

				if ( ExtraUserSlot->iSlots == 0 )
				{
					m_pExtraUserSlotList->Del(ExtraUserSlot);
				}

				res = TRUE;
			}
			else if ( ExtraUserSlot->bPermanent )
			{
				res = TRUE;
			}

			break;
		}
	}

	m_pExtraUserSlotList->UnLock();

	return res;
}

/** */
void CDownloadManager::FileListDone( CTransfer * Transfer )
{
	DCTransferQueueObject * TransferObject;
	CByteArray ba;

	if ( Transfer->GetBuffer(&ba) != 0 )
	{
		if ( (TransferObject = m_pDownloadQueue->GetUserTransferObject( Transfer->GetDstNick(), Transfer->GetHubName() )) != 0 )
		{
			TransferObject->sUserFileList.Set((char*)ba.Data(),ba.Size());

			CMessageDMFileListObject * flo = new CMessageDMFileListObject();

			flo->sNick         = TransferObject->sNick;
			flo->sHubName      = TransferObject->sHubName;
			flo->sHubHost      = TransferObject->sHubHost;
			flo->sUserFileList = TransferObject->sUserFileList;

			flo->m_eType = DC_MESSAGE_FILELIST_OBJECT;

			LogThread.Lock();

			if ( DC_DownloadManagerCallBack(flo) == -1 )
				delete flo;

			LogThread.UnLock();
		}
	}
}

/** */
bool CDownloadManager::CheckHash( CTransfer * Transfer )
{
	bool res = FALSE;
	CByteArray ba;
	DCTransferFileObject * TransferFileObject;
	DCFileChunkObject * FileChunkObject;
	CMD5 context;

	if ( (TransferFileObject = m_pDownloadQueue->GetUserFileObject( Transfer->GetDstNick(), Transfer->GetHubName(), Transfer->GetDstFilename() )) != 0 )
	{
		if ( Transfer->GetBuffer(&ba) != 0 )
		{
			if ( (TransferFileObject->m_stHash == "") && (TransferFileObject->m_bMulti == TRUE) )
			{
				// calc hash
				context.update( ba.Data(), ba.Size() );
				context.finalize();

				TransferFileObject->m_stHash = context.hex_digest();

				printf("hash is :'%s'\n",TransferFileObject->m_stHash.Data());

				m_pDownloadQueue->pChunkList->Lock();

				if ( m_pDownloadQueue->pChunkList->Get( Transfer->GetSrcFilename(), (CObject *&)FileChunkObject ) == 0 )
				{
					if ( FileChunkObject->stHash == "" )
					{
						printf("Set hash ...\n");
						FileChunkObject->stHash = TransferFileObject->m_stHash;

						res = TRUE;
					}
					else if ( FileChunkObject->stHash == TransferFileObject->m_stHash )
					{
						printf("Hash ok...\n");

						res = TRUE;
					}
					else
					{
						printf("Wrong hash !!!\n");
						TransferFileObject->m_eState = etfsERROR;
					}
				}
				else
				{
					printf("warning file chunk object not found\n");
				}

				m_pDownloadQueue->pChunkList->UnLock();
			}
			else
			{
				printf("warning hash not empty or no multi download\n");
			}
		}
		else
		{
			printf("warning file object not found\n");
		}
	}
	else
	{
		printf("warning get buffer error\n");
	}

	return res;
}

/** 
bool CDownloadManager::SetMultiDownload( CString nick, CString hubname, CString remotefile )
{
	Lock();
	bool res = FALSE;
	DCTransferQueueObject * TransferObject;
	DCTransferFileObject * TransferFileObject;
	DCFileChunkObject * FileChunkObject;

	if ( (TransferObject = m_pDownloadQueue->GetUserTransferObject( nick, hubname )) != 0 )
	{
		UserFileInfo->eWaitState    = TransferObject->eState;
		UserFileInfo->sUserFileList = TransferObject->sUserFileList;

		if ( TransferObject->pTransferFileList.Get( remotefile, (CObject *&) TransferFileObject ) == 0 )
		{
			if ( (TransferFileObject->eTransferFileState != etfsTRANSFER) &&
			     (TransferFileObject->bMulti == FALSE) )
			{
				m_pDownloadQueue->pChunkList->Lock();

				if ( (FileChunkObject = m_pDownloadQueue->GetFileChunkObject(TransferFileObject->sLocalFile)) != 0 )
				{
					DCChunkObject * ChunkObject = 0;
					while( (ChunkObject=FileChunkObject->pChunks.Next(ChunkObject)) != 0 )
					{
						if ( ChunkObject->lStart <= TEST_CHUNK_SIZE )
						{
							// error
							break;
						}
					}

					if ( ChunkObject == 0 )
				}

				m_pDownloadQueue->pChunkList->UnLock();
			}

			UserFileInfo->eFileState = TransferFileObject->eState;
			UserFileInfo->sLocalFile = TransferFileObject->sLocalFile;
			UserFileInfo->bMulti     = TransferFileObject->bMulti;
			res = TRUE;
		}
	}

	UnLock();

	return res;
}
*/

/** */
CString CDownloadManager::GetNewTransferID()
{
	lTransferID++;

	if (lTransferID==0)
		lTransferID++;

	return CString().setNum(lTransferID);
}

/** */
bool CDownloadManager::RemoveQueueFile( CString localfile )
{
	bool res = FALSE;
	CStringList * StringList;
	DCTransferQueueObject * TransferObject;
	DCTransferFileObject * TransferFileObject;

	// remove file from the chunk list
	m_pDownloadQueue->pChunkList->Lock();
	m_pDownloadQueue->pChunkList->Del( localfile );
	m_pDownloadQueue->pChunkList->UnLock();

	// remove file from the queue
	StringList = 0;

	while( m_pDownloadQueue->pQueue->Next( (CObject *&) StringList) )
	{
		TransferObject = 0;

		while( StringList->Next( (CObject *&) TransferObject) )
		{
			TransferFileObject = 0;

			while( TransferObject->pTransferFileList.Next( (CObject *&) TransferFileObject) )
			{
				if ( TransferFileObject->m_sLocalFile == localfile )
				{
					// on file transfer don't remove the file
					if ( TransferFileObject->m_eState != etfsTRANSFER )
					{
						SendFileInfo( TransferObject, TransferFileObject, TRUE );
						TransferObject->pTransferFileList.Del(TransferFileObject->m_sRemoteFile);
						TransferFileObject = 0;
						res = TRUE;
					}
					else
					{
						printf("WARNING: RemoveQueueFile: file transfer is running\n");
					}
				}
			}
		}
	}

	return res;
}

/** */
bool CDownloadManager::RemoveQueueFile( CString nick, CString hubname, CString remotefile )
{
	bool res = FALSE;
	DCTransferFileObject * TransferFileObject = 0;
	DCTransferQueueObject * TransferObject = 0;

	if ( (TransferObject = m_pDownloadQueue->GetUserTransferObject( nick, hubname )) != 0 )
	{
		if ( remotefile != "" )
		{
			if ( (TransferFileObject = m_pDownloadQueue->GetUserFileObject( nick, hubname, remotefile )) != 0 )
			{
				if ( TransferFileObject->m_eState != etfsTRANSFER )
				{
					res = TRUE;
				}
				else
				{
					printf("WARNING: RemoveQueueFile: file transfer is running\n");
				}
			}
		}
		else
		{
			if ( TransferObject->eState != etwsRUN )
			{
				res = TRUE;
			}
			else
			{
				printf("WARNING: RemoveQueueFile: transfer is running\n");
			}
		}

		if ( res == TRUE )
		{
			SendFileInfo( TransferObject, TransferFileObject, TRUE );
			res = m_pDownloadQueue->DelUserFileObject( nick, hubname, remotefile );
		}
	}

	return res;
}

/** */
bool CDownloadManager::CheckWaitTransfer( CTransfer * Transfer )
{
	m_pDownloadQueue->pQueue->Lock();

	bool res = FALSE;
	DCTransferQueueObject * TransferObject;
	DCTransferWait * TransferWait = 0;
	DCTransferBanObject * TransferBanObject;
	CString host;
	int port;

	printf("CheckWaitTransfer I : %s on %s\n",
		Transfer->GetDstNick().Data(),
		Transfer->GetHubName().Data() );

	// allow only 1 request in 30 seconds
	m_pTransferBanList->Lock();

	TransferBanObject = 0;

	if ( Transfer->GetPeerName( &host, &port ) == FALSE )
	{
		// nothing ...
	}
	else if ( m_pTransferBanList->Get( host, (CObject*&)TransferBanObject ) != 0 )
	{
		TransferBanObject = new DCTransferBanObject();

		TransferBanObject->m_sIP   = host;
		TransferBanObject->m_tTime = time(0);

		m_pTransferBanList->Add( Transfer->GetDstNick(), TransferBanObject );
	}
	else
	{
		TransferBanObject->m_nRequestCount++;
	}

	if ( TransferBanObject->m_nRequestCount > 0 )
	{
		if ( ((time(0)-TransferBanObject->m_tTime)/TransferBanObject->m_nRequestCount) > 30 )
		{
			Transfer->Disconnect(TRUE);
			SendLogInfo( "WARNING: Disconnect aggressive client " + host );
			TransferBanObject = 0;
		}
	}

	m_pTransferBanList->UnLock();

	if ( TransferBanObject == 0 )
	{
		m_pDownloadQueue->pQueue->UnLock();
		return res;
	}

	// search in the wait list ...
	m_pTransferWaitList->Lock();

	while ( (TransferWait = m_pTransferWaitList->Next(TransferWait)) != 0 )
	{
		// search for the nick ...
		if ( ((TransferWait->sNick == Transfer->GetDstNick()) && (TransferWait->sNick != "")) ||
		     ((TransferWait->sUserHost == Transfer->GetHost()) && (TransferWait->sUserHost != "")) )
		{
			Transfer->SetHubName(TransferWait->sHubName);
			Transfer->SetHubHost(TransferWait->sHubHost);

			break;
		}
	}

	m_pTransferWaitList->UnLock();

	printf("CheckWaitTransfer II: %s on %s [%d]\n",
		Transfer->GetDstNick().Data(),
		Transfer->GetHubName().Data(),(TransferWait!=0) );

	if ( (TransferObject = m_pDownloadQueue->GetUserTransferObject( Transfer->GetDstNick(), Transfer->GetHubName() )) != 0 )
	{
		printf("Waiting: %s on %s %s\n",
			TransferObject->sNick.Data(),
			TransferObject->sHubName.Data(),
			TransferObject->sHubHost.Data() );

		if ( TransferObject->eState == etwsIDLE )
		{
			SendLogInfo( "WARNING: Increase the response timeout." );
		}

		if ( (TransferObject->eState == etwsWAIT) || (TransferObject->eState == etwsIDLE) )
		{
			printf("wait found ...\n");

			TransferObject->eState = etwsRUN;
			TransferObject->iConnections++;

			// waiting transfer found ...
			res = TRUE;

			SendFileInfo( TransferObject );
		}
		else
		{
			printf("ERROR: wait in wrong state (please report!) (%d/%d)\n",TransferObject->eState,TransferObject->iConnections);
		}
	}

	if ( (res == FALSE) && (TransferWait == 0) )
	{
		printf("Warning: no wait transfer found for '%s'\n",Transfer->GetDstNick().Data());
		Transfer->Disconnect(TRUE);
	}

	if ( TransferWait != 0 )
	{
		m_pTransferWaitList->Lock();
		m_pTransferWaitList->Del(TransferWait);
		m_pTransferWaitList->UnLock();
	}

	m_pDownloadQueue->pQueue->UnLock();

	return res;
}

/** */
bool CDownloadManager::ChangeDirection( CTransfer * Transfer )
{
	bool res = FALSE;
	DCTransferQueueObject * TransferObject;

	m_pDownloadQueue->pQueue->Lock();

	if ( (TransferObject = m_pDownloadQueue->GetUserTransferObject( Transfer->GetDstNick(), Transfer->GetHubName() )) != 0 )
	{
		printf("Waiting: %s on %s %s\n",
			TransferObject->sNick.Data(),
			TransferObject->sHubName.Data(),
			TransferObject->sHubHost.Data() );

		// change to upload ...
		if ( Transfer->GetSrcDirection() == edDOWNLOAD )
		{
			if ( TransferObject->eState == etwsRUN )
			{
				TransferObject->iConnections--;
				if ( TransferObject->iConnections == 0 )
					TransferObject->eState = etwsIDLE;
				SendFileInfo( TransferObject );

				printf("change transfer -> upload ...\n");

				res = TRUE;
			}
			else
			{
				printf("can't change transfer upload ...\n");
			}
		}
	}

	m_pDownloadQueue->pQueue->UnLock();

	return res;
}

/** */
bool CDownloadManager::SetNextFile( CTransfer * Transfer )
{
	bool res;

	m_pDownloadQueue->pQueue->Lock();
	res = SetFile(Transfer);
	m_pDownloadQueue->pQueue->UnLock();

	return res;
}

/** */
bool CDownloadManager::SetDirection( CTransfer * Transfer )
{
	CTransferObject * TransferObject;
	CObject *obj;
	int i;
	bool res = FALSE;

	// check max uploads for a user
	if ( Transfer->GetSrcDirection() == edUPLOAD )
	{
		// check for max user upload slots
		obj = 0;
		i = 0;

		// the list is allready locked by the thread
		while( m_pTransferList->Next( (CObject*&)obj) )
		{
			TransferObject = (CTransferObject*)obj;

			if ( TransferObject->m_pTransfer->GetDstDirection() == edDOWNLOAD )
			{
				if ( Transfer->GetDstNick() == TransferObject->m_pTransfer->GetDstNick() )
				{
					i++;
				}
			}
		}

		// the current transfer is allready in the list
		if ( (CConfig::Instance()->GetUserUploadSlots() == 0) ||
		     (CConfig::Instance()->GetUserUploadSlots() >= i) )
		{
			// check for extra user slots
			if ( (res = CheckUserSlot( Transfer->GetDstNick(), Transfer->GetHubName() )) == TRUE )
			{
				DownloadManagerInfo.slot_use_user++;
				Transfer->SetTransferType(ettUSER);
				res = TRUE;
			}
			else
			{
				// check slot limit
				i = CConfig::Instance()->GetMaxUpload();

				// unlimit
				if ( i == 0 )
				{
					res = TRUE;
				}
				else if ( i > DownloadManagerInfo.slot_use_settings )
				{
					res = TRUE;
				}

				if (res)
				{
					DownloadManagerInfo.slot_use_settings++;
					Transfer->SetTransferType(ettSETTINGS);
				}
				else // no more free slots ... a op can download the filelist ....
        			{
					if ( Transfer->GetDstNick() == "" )
					{
						printf("WARNING: get a free slot -> remote nick is empty\n");
					}
					else if ( CServerManager::Instance()->IsAdmin(Transfer->GetHubName(),Transfer->GetDstNick()) == TRUE )
					{
						if ( 3 >= DownloadManagerInfo.slot_use_operator )
						{
							DownloadManagerInfo.slot_use_operator++;
							Transfer->SetTransferType(ettOPERATOR);
							res = TRUE;
						}
					}
					else
					{
						// is not a operator ... set to special transfer
						if ( 3 >= DownloadManagerInfo.slot_use_special )
						{
							DownloadManagerInfo.slot_use_special++;
							Transfer->SetTransferType(ettSPECIAL);
							res = TRUE;
						}
					}
				}
			}
		}
	}
	// allow all downloads
	else
	{
		res = TRUE;
	}

	return res;
}

/** */
void CDownloadManager::OptimizeChunks( DCFileChunkObject * FileChunkObject )
{
	DCChunkObject *co, *co1;

	// get first not locked chunk
	co = 0;
	while ( (co=FileChunkObject->pChunks.Next(co)) != 0 )
	{
		if ( co->bLocked == FALSE )
		{
			co1 = co;
			while ( (co1=FileChunkObject->pChunks.Next(co1)) != 0 )
			{
				if ( co1 == co )
				{
					continue;
				}

				if ( co1->bLocked == FALSE )
				{
					if ( co->lEnd == co1->lStart )
					{
						co->lEnd = co1->lEnd;
						FileChunkObject->pChunks.Del(co1);
						co1 = co;
					}
					else if ( co->lStart == co1->lEnd )
					{
						co->lStart = co1->lStart;
						FileChunkObject->pChunks.Del(co1);
						co1 = co;
					}
				}
			}
		}
	}

}

/** */
bool CDownloadManager::GetNextChunk( CString file, ulonglong * lstart, ulonglong * lend )
{
	bool res = FALSE;
	DCFileChunkObject * FileChunkObject;
	DCChunkObject * ChunkObject, *co, *co1;
	ulonglong size;
	bool split;

	printf("get the next chunk for '%s'\n",file.Data());

	if ( (FileChunkObject = m_pDownloadQueue->GetFileChunkObject(file)) != 0 )
	{
		OptimizeChunks(FileChunkObject);

		co = 0;
		ChunkObject = 0;
		size = 0;

		// get a big free chunk
		while ( (ChunkObject=FileChunkObject->pChunks.Next(ChunkObject)) != 0 )
		{
			if ( ChunkObject->bLocked == FALSE )
			{
				if ( (ChunkObject->lEnd-ChunkObject->lStart) > size )
				{
					co = ChunkObject;
					size = ChunkObject->lEnd-ChunkObject->lStart;
				}
			}
		}

		// found a free chunk
		if ( co != 0 )
		{
			split = FALSE;
			ChunkObject = 0;

			// check if a running chunk for this chunk
			while ( (ChunkObject=FileChunkObject->pChunks.Next(ChunkObject)) != 0 )
			{
				if ( ChunkObject->bLocked == TRUE )
				{
					if ( ChunkObject->lEnd == co->lStart )
					{
						split = TRUE;
					}
				}
			}

			ChunkObject = co;

			if ( (ChunkObject->lEnd - ChunkObject->lStart) > CHUNK_SIZE )
			{
				// create a new chunk
				co = new DCChunkObject();
	
				// split chunk
				if ( (FileChunkObject->bMulti == FALSE) || (ChunkObject->lStart == 0) || (split == FALSE) )
				{
					co->lStart  = ChunkObject->lStart;
					co->lEnd    = ChunkObject->lStart+CHUNK_SIZE;
					ChunkObject->lStart += CHUNK_SIZE;
				}
				else
				{
					printf("CHUNK SET 1: %llu %llu\n",ChunkObject->lStart,ChunkObject->lEnd);

					co->lStart = ChunkObject->lStart+((ChunkObject->lEnd - ChunkObject->lStart)/2);

					if ( (ChunkObject->lEnd-co->lStart) > CHUNK_SIZE )
					{
						co->lEnd = co->lStart+CHUNK_SIZE;
						co1 = new DCChunkObject();
						co1->lStart = co->lEnd;
						co1->lEnd   = ChunkObject->lEnd;
						FileChunkObject->pChunks.Add(co1);
						printf("CHUNK SET 2: %llu %llu\n",co1->lStart,co1->lEnd);
					}
					else
					{
						co->lEnd = ChunkObject->lEnd;
					}

					ChunkObject->lEnd = co->lStart;

					printf("CHUNK SET 3: %llu %llu\n",ChunkObject->lStart,ChunkObject->lEnd);
					printf("CHUNK SET 4: %llu %llu\n",co->lStart,co->lEnd);
				}

				co->bLocked = TRUE;

				FileChunkObject->pChunks.Add(co);

				*lstart = co->lStart;
				*lend   = co->lEnd;

				printf("NEW CHUNK SPLIT/LOCKED: %llu %llu\n",*lstart,*lend);

				res = TRUE;
			}
			else
			{
				ChunkObject->bLocked = TRUE;

				*lstart = ChunkObject->lStart;
				*lend   = ChunkObject->lEnd;

				printf("NEW CHUNK LOCKED: %llu %llu\n",*lstart,*lend);

				res = TRUE;
			}
		}
	}
	else
	{
		printf("warning file not found in the chunk list\n");
	}

	return res;
}

/** 0: not found; 1: found and update; 2: no more chunks (file download ok) */
int CDownloadManager::UpdateChunk( CString file, ulonglong lstart, ulonglong lend, ulonglong lcurrent )
{
	int res = 0;
	DCFileChunkObject * FileChunkObject;
	DCChunkObject * ChunkObject, *co;

	printf("update chunk for '%s'\n",file.Data());

	m_pDownloadQueue->pChunkList->Lock();

	if ( (FileChunkObject = m_pDownloadQueue->GetFileChunkObject(file)) != 0 )
	{
		ChunkObject = 0;

		// search the chunk
		while ( (ChunkObject=FileChunkObject->pChunks.Next(ChunkObject)) != 0 )
		{
			if ( (ChunkObject->lStart == lstart) && (ChunkObject->lEnd == lend) )
			{
				// chunk found
				res = 1;

				if ( ChunkObject->bLocked == FALSE )
				{
					printf("warning wrong chunk state\n");
				}

				ChunkObject->bLocked = FALSE;

				if ( lstart != lcurrent )
				{
					// update size done
					FileChunkObject->lSizeDone += (lcurrent-lstart);

					printf("FILESTATE: %llu %llu\n",FileChunkObject->lSizeDone,FileChunkObject->lSize);

					if ( lcurrent == lend )
					{
						FileChunkObject->pChunks.Del(ChunkObject);
						ChunkObject = 0;
					}
					else
					{
						ChunkObject->lStart = lcurrent;
					}

					if ( FileChunkObject->lSizeDone == FileChunkObject->lSize )
					{
						// remove the complete chunk from the list
						m_pDownloadQueue->pChunkList->Del(file);
						res = 2;
						break;
					}
				}

				if ( ChunkObject != 0 )
				{
					co = 0;
					while ( (co=FileChunkObject->pChunks.Next(co)) != 0 )
					{
						if ( co == ChunkObject )
							continue;

						if ( ChunkObject->lEnd == co->lStart )
						{
							if ( co->bLocked == FALSE )
							{
								co->lStart = ChunkObject->lStart;
								FileChunkObject->pChunks.Del(ChunkObject);
								ChunkObject = co;
								printf("CHUNK FIX1: %llu %llu\n",ChunkObject->lStart,ChunkObject->lEnd);
							}
							break;
						}
					}

					co = 0;
					while ( (co=FileChunkObject->pChunks.Next(co)) != 0 )
					{
						if ( co == ChunkObject )
							continue;

						if ( ChunkObject->lStart == co->lEnd )
						{
							if ( co->bLocked == FALSE )
							{
								co->lEnd = ChunkObject->lEnd;
								FileChunkObject->pChunks.Del(ChunkObject);
								ChunkObject = 0;
								printf("CHUNK FIX2: %llu %llu\n",co->lStart,co->lEnd);
							}
							break;
						}
					}
				}

				break;
			}
		}
	}
	else
	{
		printf("warning file not found in the chunk list\n");
	}

	m_pDownloadQueue->pChunkList->UnLock();

	return res;
}

/** */
bool CDownloadManager::GetNewChunkEnd( CString file, ulonglong lstart, ulonglong lend, ulonglong lcurrent, ulonglong * lnstart, ulonglong *lnend )
{
	m_pDownloadQueue->pChunkList->Lock();

	bool res = FALSE;
	DCFileChunkObject * FileChunkObject;
	DCChunkObject * ChunkObject, *co1=0,*co2=0;
	ulonglong chunk_size;

	if ( (FileChunkObject = m_pDownloadQueue->GetFileChunkObject(file)) != 0 )
	{
		ChunkObject = 0;

		while ( (ChunkObject=FileChunkObject->pChunks.Next(ChunkObject)) != 0 )
		{
			if ( (ChunkObject->lStart == lstart) && (ChunkObject->lEnd == lend) )
			{
				// start chunk found
				co1 = ChunkObject;

				if ( co2 )
				{
					// next chunk is set -> stop search
					break;
				}
			}
			else if ( ChunkObject->lStart == lend )
			{
				// next chunk found
				if ( ChunkObject->bLocked == TRUE )
				{
					// the chunk after the original chunk is locked -> stop search
					break;
				}
				else
				{
					// set next chunk
					co2 = ChunkObject;
       	
					if ( co1 )
					{
						// start chunk is set -> stop search
						break;
					}
				}
			}
		}

		// found start and next chunk
		if ( (co1 != 0) && (co2 != 0) )
		{
			printf("set new chunk end for '%s'\n",file.Data());

			// calc chunk size
			if ( (lend-lcurrent) > CHUNK_SIZE )
				chunk_size = CHUNK_SIZE;
			else
				chunk_size = CHUNK_SIZE-(lend-lcurrent);

			// if chunk > max chunk split chunk to chunk size
			if ( (co2->lEnd-co2->lStart) > chunk_size )
			{
				// set new chunk end
				co1->lEnd   += chunk_size;
				// set correct new chunk start
				co2->lStart += chunk_size;
			}
			else
			{
				// set new chunk end
				co1->lEnd = co2->lEnd;
				// remote the chunk
				FileChunkObject->pChunks.Del(co2);
			}

			if ( (lcurrent-lstart) > 0 )
			{
				FileChunkObject->lSizeDone += (lcurrent-lstart);
				co1->lStart = lcurrent;
			}

			*lnstart = co1->lStart;
			*lnend   = co1->lEnd;

			printf("new chunk end set %llu -> %llu [%llu/%llu]\n",lend,*lnend,(*lnend)-(*lnstart),chunk_size);

			res = TRUE;
		}
	}
	else
	{
		printf("warning file not found in the chunk list\n");
	}

	m_pDownloadQueue->pChunkList->UnLock();

	return res;
}

/** */
bool CDownloadManager::SetFile( CTransfer * Transfer )
{
	bool res;
	DCTransferQueueObject * TransferObject;
	ulonglong lstart,lend;
	int priority = 0;
	res = FALSE;

	if ( (TransferObject = m_pDownloadQueue->GetUserTransferObject( Transfer->GetDstNick(), Transfer->GetHubName() )) != 0 )
	{
		if ( TransferObject->pTransferFileList.Count() > 0 )
		{
			DCTransferFileObject * TransferFileObject = 0;

			while( (priority <= MAX_FILE_PRIORITY) && (TransferFileObject == 0) )
			{
			while( TransferObject->pTransferFileList.Next((CObject*&)TransferFileObject) )
			{
				// check the priority
				if ( priority != TransferFileObject->m_nPriority )
					continue;

				if ( TransferFileObject->m_eState == etfsNONE )
				{
					printf("set file: '%s'\n",TransferFileObject->m_sRemoteFile.Data());

					CString sPath="",sLocalPath="",sFile="";
					CDir dir;

					if ( TransferFileObject->m_eMedium == eltFILE )
					{
						sFile = TransferFileObject->m_sLocalFile;

						printf("DEBUG: file: '%s'\n",sFile.Data());

						// extract file
						int pos = sFile.FindRev(DIRSEPARATOR);
						if( pos != -1 )
						{
							sPath = sFile.Left(pos);
						}

						printf("DEBUG: path: '%s'\n", sPath.Data());

						// create the path
						if ( dir.CreatePath(sPath) == FALSE )
						{
							TransferFileObject->m_eState = etfsERROR;
							SendFileInfo( TransferObject, TransferFileObject );
							SendLogInfo( "Create path failed: " + sPath, Transfer );
							printf("DEBUG: create path failed: '%s'\n", sPath.Data());
						}
						else
						{
							printf("DOWNLOAD: '%s' %llu '%s'\n",
								TransferFileObject->m_sRemoteFile.Data(),
								TransferFileObject->m_nSize,
								sFile.Data());

								res = TRUE;
						}
					}
					else
					{
						res = TRUE;
					}

					if ( res == TRUE )
					{
						// first we create the hash over the first 1MB of a multi download file in a buffer
						if ( (TransferFileObject->m_bMulti == TRUE) &&
						     (TransferFileObject->m_stHash == ""))
						{
							printf("create the hash for the file\n");

							Transfer->SetMedium(eltBUFFER);

							lstart = 0;
							lend   = TEST_CHUNK_SIZE;
						}
						else // get the next chunk of the file
						{
							if ( TransferFileObject->m_sRemoteFile == CString(DC_USER_FILELIST) )
							{
								lstart = 0;
								lend   = 0;
							}
							else if ( GetNextChunk( TransferFileObject->m_sLocalFile, &lstart, &lend ) == FALSE )
							{
								// no more chunks
								printf("no more chunks ...\n");
								continue;
							}

							Transfer->SetMedium(TransferFileObject->m_eMedium);
						}

						Transfer->SetDone(etsNONE);

						// mark file as transfered
						TransferFileObject->m_eState = etfsTRANSFER;

						Transfer->StartDownload(TransferFileObject->m_sRemoteFile,lstart,lstart,lend,TransferFileObject->m_nSize,lend-lstart,sFile);
					}
					else
					{
						continue;
					}

					SendFileInfo( TransferObject, TransferFileObject );

					break;
				}
				else
				{
//					printf("state file: '%s' %d\n",TransferFileItem->sRemoteName.Data(),TransferFileItem->eState);
				}
			}

			priority++;
			}
		}
	}

	if ( res == FALSE )
	{
		// set disconnect timeout ... save remote slot ;-)
		if ( Transfer->GetDone() != etsIDLE )
		{
			Transfer->SetStartTime(time(0));
			Transfer->SetDone(etsIDLE);
		}
	}

	return res;
}

/** */
void CDownloadManager::UpdateFileState( CTransfer * Transfer, eTransferFileState eState )
{
	DCTransferQueueObject * TransferObject;
	DCTransferFileObject * TransferFileObject;

	m_pDownloadQueue->pQueue->Lock();

	printf("updatefile\n");
	if ( (TransferObject = m_pDownloadQueue->GetUserTransferObject( Transfer->GetDstNick(), Transfer->GetHubName() )) != 0 )
	{
		TransferFileObject = m_pDownloadQueue->GetUserFileObject( Transfer->GetDstNick(), Transfer->GetHubName(), Transfer->GetDstFilename() );

		if ( TransferFileObject != 0 )
		{
			if ( TransferFileObject->m_eState != etfsTRANSFER )
			{
				printf("warning, wrong state in updatefile\n");
			}
			// if this a check hash dl ?
			else if ( (Transfer->GetMedium() == eltBUFFER) && (TransferFileObject->m_bMulti == TRUE) )
			{
				printf("updatefile hash\n");
				if ( eState == etfsNONE )
				{
					if ( Transfer->GetCurrentPosition() == Transfer->GetEndPosition() )
					{
						if ( CheckHash(Transfer) == TRUE )
						{
							SendLogInfo( "Hash ok '" + TransferFileObject->m_sRemoteFile + "'", Transfer );
							TransferFileObject->m_eState = etfsNONE;
							// wan't a reconnect
							TransferObject->bReconnect = TRUE;
						}
						else
						{
							SendLogInfo( "Hash failed '" + TransferFileObject->m_sRemoteFile + "'", Transfer );
							TransferFileObject->m_eState = etfsERROR;
						}
					}
					else
					{
						TransferFileObject->m_eState = eState;
					}
				}
				else
				{
					TransferFileObject->m_eState = eState;
				}

				SendFileInfo( TransferObject, TransferFileObject );
			}
			else
			{
				printf("updatefile normal\n");

				int state=0;

				TransferFileObject->m_eState = eState;

				// check if mylist.dclist done
				if ( TransferFileObject->m_sRemoteFile == CString(DC_USER_FILELIST) )
				{
					if ( (Transfer->GetLength() != 0) &&
					     (Transfer->GetLength() == Transfer->GetCurrentPosition()) )
					{
						state = 2;
					}
				}
				else
				{
					state = UpdateChunk( TransferFileObject->m_sLocalFile, Transfer->GetStartPosition(),
						Transfer->GetEndPosition(), Transfer->GetCurrentPosition() );
				}

				TransferFileObject->m_nSize = Transfer->GetLength();

				// download finished
				if ( state == 2 )
				{
					SendLogInfo( "Transfer done '" + TransferFileObject->m_sRemoteFile + "'", Transfer );
					SendFileInfo( TransferObject, TransferFileObject, TRUE );
					SendTransferInfo( Transfer, FALSE );

					// write in logfile
					if ( (TransferFileObject->m_eMedium == eltFILE) &&
						(CConfig::Instance()->GetLogFile() == TRUE) &&
						(TransferFileObject->m_sRemoteFile != CString(DC_USER_FILELIST)) )
					{
						CLogFile::Write(CConfig::Instance()->GetLogFileName(),eltINFO,"Transfer done '" + TransferFileObject->m_sLocalFile + "'");
					}
					
					// move the file to the finished path
					if ( (TransferFileObject->m_eMedium == eltFILE) && // only files
					     (CConfig::Instance()->GetDownloadFinishedFolder() != "") && // only if dlfolder set
					     (TransferFileObject->m_sRemoteFile != CString(DC_USER_FILELIST)) && // only if this not a filelist
					     (TransferFileObject->m_sLocalFile.Find(CConfig::Instance()->GetDownloadFolder().Data()) == 0) ) // only if the download in the download folder
					{
						CString s;

						CDir dir(CConfig::Instance()->GetDownloadFinishedFolder());

						if ( dir.CreatePath(TransferFileObject->m_sLocalPath) == TRUE )
						{
							dir.SetPath(CConfig::Instance()->GetDownloadFinishedFolder()+DIRSEPARATOR+TransferFileObject->m_sLocalPath);
#ifdef WIN32
							s  = "move ";
							s += "'" + TransferFileObject->m_sLocalFile + "' ";
							s += "'" + dir.Path() + "'";

							printf("move file '%s'\n",s.Data());

							if ( system(s.Data()) != 0 )
							{
								printf("move failed !\n");
							}
#else
							s = dir.Path() +DIRSEPARATOR+ TransferFileObject->m_sLocalFileName;
							printf("move file: '%s' ---> '%s'\n",TransferFileObject->m_sLocalFile.Data(), s.Data());
							if ( rename(TransferFileObject->m_sLocalFile.Data(), s.Data()) != 0 )
							{
								if (errno == EXDEV)
								{
									// we have to copy it manually
									if (fastcopy(TransferFileObject->m_sLocalFile.Data(), s.Data()) == FALSE)
									{
										printf("move failed !\n");
									}
									else
									{
										unlink(TransferFileObject->m_sLocalFile.Data());
									}
								}
							}
#endif
						}
						else
						{
							printf("move failed (create path)!\n");
						}
					}

					if ( TransferFileObject->m_bMulti == TRUE )
					{
						// remove all files from the wait queue
						RemoveQueueFile(TransferFileObject->m_sLocalFile);
					}
					else
					{
						// remove a userlist only from this user/hub
						RemoveQueueFile( TransferObject->sNick, TransferObject->sHubName, TransferFileObject->m_sRemoteFile );
					}
				}
				else
				{
					SendFileInfo( TransferObject, TransferFileObject );
				}
			}
		}
	}

	m_pDownloadQueue->pQueue->UnLock();
}

/** */
void CDownloadManager::UpdateBanList( time_t ttimeout )
{
	CString s;
	DCTransferBanObject * TransferBanObject = 0, * tbo = 0;

	// remove old entrys
	m_pTransferBanList->Lock();

	while ( m_pTransferBanList->Next( s, (CObject*&)TransferBanObject ) == 1 )
	{
		if ( (ttimeout-TransferBanObject->m_tTime) > 180 )
		{
			m_pTransferBanList->Del(s);
			TransferBanObject = tbo;
		}
		else
		{
			tbo = TransferBanObject;
		}
	}

	m_pTransferBanList->UnLock();
}

/** */
void CDownloadManager::UpdateTransferList( time_t ttimeout )
{
	long count,rate,tmprate;
	CTransferObject * TransferObject;
	CObject *obj,*oobj;

	m_pTransferList->Lock();

	count = m_pTransferList->Count();

	// check if shutdown ready
	if ( (m_eShutdownState == essSHUTDOWN) && (count == 0) )
	{
		m_eShutdownState = essSHUTDOWNREADY;
	}

	if ( (ttimeout-tUpdateTransferTimeout) >= 1 )
	{
		// update slots
		DownloadManagerInfo.slot_max = CConfig::Instance()->GetMaxUpload();
		DownloadManagerInfo.rate_ul_settings = 0;
		DownloadManagerInfo.rate_ul_operator = 0;
		DownloadManagerInfo.rate_ul_user     = 0;
		DownloadManagerInfo.rate_ul_special  = 0;
		DownloadManagerInfo.rate_dl          = 0;

		UpdateBanList( ttimeout );
	}

	if ( count > 0 )
	{
		obj = oobj = 0;

		while( m_pTransferList->Next( (CObject*&)obj) )
		{
			TransferObject = (CTransferObject*)obj;

			// check if user offline and disconnect
			if ( (ttimeout-TransferObject->m_UserDisconnectTimeout) > 180 )
			{
				eCloseType closetype = CConfig::Instance()->GetHubOfflineTransferClose();

				TransferObject->m_UserDisconnectTimeout = ttimeout;

				if ( CServerManager::Instance()->IsUserOnline( TransferObject->m_pTransfer->GetDstNick(), TransferObject->m_pTransfer->GetHubName(), 0 ) == FALSE )
				{
					if ( (TransferObject->m_pTransfer->GetSrcDirection()==edDOWNLOAD) &&
					   ( (closetype==ectBOTH) || (closetype==ectDLD) ) )
					{
						TransferObject->m_pTransfer->Disconnect(TRUE);
						SendLogInfo("Disconnect offline user: "+TransferObject->m_pTransfer->GetDstNick()+"@"+TransferObject->m_pTransfer->GetHubName());
					}
					else if ( (TransferObject->m_pTransfer->GetSrcDirection()==edUPLOAD) &&
					        ( (closetype==ectBOTH) || (closetype==ectUPLD) ) )
					{
						TransferObject->m_pTransfer->Disconnect(TRUE);
						SendLogInfo("Disconnect offline user: "+TransferObject->m_pTransfer->GetDstNick()+"@"+TransferObject->m_pTransfer->GetHubName());
					}
				}
			}

			// refresh transfer
			TransferObject->m_pTransfer->Thread(0);

			if ( (ttimeout-tUpdateTransferTimeout) >= 1 )
			{
				// done==2 set new file or disconnect
				if ( (TransferObject->m_pTransfer->GetDone() == etsIDLE) && (m_eShutdownState == essNONE) )
				{
					if ( SetFile(TransferObject->m_pTransfer) == FALSE )
					{
						// no new file ...
						if ( TransferObject->m_pTransfer->GetStartTime() != 0 )
						{
							if ( (ttimeout-TransferObject->m_pTransfer->GetStartTime()) >= 60 )
							{
								TransferObject->m_pTransfer->SetStartTime(0);
								TransferObject->m_pTransfer->Disconnect(TRUE);
							}
						}
					}

					if ( (ttimeout-tUpdateTransferTimeout) >= 1 )
					{
						SendTransferInfo(TransferObject->m_pTransfer);
					}

					oobj = obj;
				}
				else if ( TransferObject->m_pTransfer->GetDone() == etsREADY )
				{
					if ( TransferObject->m_pTransfer->GetDstDirection() == edDOWNLOAD )
					{
						// TODO: update slot infos etc. in downloadmanager and gui
						if ( TransferObject->m_pTransfer->GetTransferType() == ettSETTINGS )
						{
							if ( DownloadManagerInfo.slot_use_settings > 0 )
								DownloadManagerInfo.slot_use_settings--;
						}
						else if ( TransferObject->m_pTransfer->GetTransferType() == ettOPERATOR )
						{
							if ( DownloadManagerInfo.slot_use_operator > 0 )
								DownloadManagerInfo.slot_use_operator--;
						}
						else if ( TransferObject->m_pTransfer->GetTransferType() == ettUSER )
						{
							if ( DownloadManagerInfo.slot_use_user > 0 )
								DownloadManagerInfo.slot_use_user--;
						}
						else if ( TransferObject->m_pTransfer->GetTransferType() == ettSPECIAL )
						{
							if ( DownloadManagerInfo.slot_use_settings > 0 )
								DownloadManagerInfo.slot_use_special--;
						}

						// update used slots
						CServerManager::Instance()->SendMyInfoToConnectedServers();
					}

					SendTransferInfo( TransferObject->m_pTransfer, TRUE );

					// update internal transfer state for this user/hubname
					// set to idle (this work only with 1 transfer per user/hubname ...)
					DCTransferQueueObject * TransferQueueObject;

					if ( TransferObject->m_pTransfer->GetSrcDirection() == edDOWNLOAD )
					{
						if ( (TransferQueueObject = m_pDownloadQueue->GetUserTransferObject( TransferObject->m_pTransfer->GetDstNick(), TransferObject->m_pTransfer->GetHubName() )) != 0 )
						{
							TransferQueueObject->iConnections--;
							if ( TransferQueueObject->iConnections == 0 )
								TransferQueueObject->eState = etwsIDLE;
							// wan't reconnect ?
							if ( TransferQueueObject->bReconnect == TRUE )
							{
								TransferQueueObject->tTimeout = 0;
								TransferQueueObject->bReconnect = FALSE;
							}
							else
							{
								TransferQueueObject->tTimeout = ttimeout;
							}
							SendFileInfo( TransferQueueObject );
						}
					}

					m_pTransferList->Del( TransferObject->m_pTransfer->GetTransferID() );

					obj = oobj;
				}
				else if ( m_eShutdownState == essNONE )
				{
					if ( (ttimeout-tUpdateTransferTimeout) >= 1 )
					{
						SendTransferInfo(TransferObject->m_pTransfer);
					}

					if ( TransferObject->m_pTransfer->GetSrcDirection() == edUPLOAD )
					{
						switch(TransferObject->m_pTransfer->GetTransferType())
						{
							case ettSETTINGS:
								DownloadManagerInfo.rate_ul_settings += TransferObject->m_pTransfer->GetTransferrate();
								break;
							case ettOPERATOR:
								DownloadManagerInfo.rate_ul_operator += TransferObject->m_pTransfer->GetTransferrate();
								break;
							case ettUSER:
								DownloadManagerInfo.rate_ul_user += TransferObject->m_pTransfer->GetTransferrate();
								break;
							case ettSPECIAL:
								DownloadManagerInfo.rate_ul_special += TransferObject->m_pTransfer->GetTransferrate();
								break;
							default:
								break;
						}
					}
					else if ( TransferObject->m_pTransfer->GetSrcDirection() == edDOWNLOAD )
					{
						DownloadManagerInfo.rate_dl += TransferObject->m_pTransfer->GetTransferrate();
					}
					oobj = obj;
				}
				else
				{
					oobj = obj;
				}
			}
		}
	}


	if ( (ttimeout-tUpdateTransferTimeout) >= 1 )
	{
		// dynamic upload rate
		if ( (CConfig::Instance()->GetDynamicUploadRate() == TRUE) &&
		     (CConfig::Instance()->GetMaxUploadRate() > 0) &&
		     (DownloadManagerInfo.slot_use_settings > 0) &&
		     (DownloadManagerInfo.slot_max > 0) &&
		     (DownloadManagerInfo.rate_ul_settings<(CConfig::Instance()->GetMaxUploadRate()*DownloadManagerInfo.slot_max)) )
		{
			// calculate max rate for every slot
			rate = CConfig::Instance()->GetMaxUploadRate()*DownloadManagerInfo.slot_max;

			// this is the max rate for every slot but if a slot use less than
			// we give the other slots the "restrate"
			rate /= DownloadManagerInfo.slot_use_settings;

			if ( rate > 0 )
			{
				obj     = 0;
				count   = 0;
				tmprate = 0;

				while( m_pTransferList->Next( (CObject*&)obj) )
				{
					TransferObject = (CTransferObject*)obj;

					if ( (TransferObject->m_pTransfer->GetDone() == etsNONE) &&
					     (TransferObject->m_pTransfer->GetSrcDirection() == edUPLOAD) &&
					     (TransferObject->m_pTransfer->GetTransferType() == ettSETTINGS) )
					{
						// check if the slot use less than 1/2 of the max rate
						if ( (TransferObject->m_pTransfer->GetTransferrate() < (TransferObject->m_pTransfer->GetRate()/2)) && ((ulonglong)rate >= TransferObject->m_pTransfer->GetRate()) )
						{
							// set the new rate for this transfer (3/4)
							TransferObject->m_pTransfer->SetRate((TransferObject->m_pTransfer->GetRate()*3/4));
							// give other slots the restrate
							tmprate += rate-TransferObject->m_pTransfer->GetRate();
						}
						// check if the slot use less than 3/4 but more as 1/2 of the current rate we made no changes
						else if ( (TransferObject->m_pTransfer->GetTransferrate() < (TransferObject->m_pTransfer->GetRate()*3/4)) && ((ulonglong)rate >= TransferObject->m_pTransfer->GetRate()) )
						{
							// give other slots the restrate
							tmprate += rate-TransferObject->m_pTransfer->GetRate();
						}
						// all other slots use more as 3/4 or the max rate for a slot is less than the current rate for the transfer
						else
						{
							count++;
							tmprate += rate;
						}
					}
				}

				// calc new rate
				if ( (tmprate > 0) && (count > 0) )
				{
					tmprate /= count;
					obj     = 0;

					while( m_pTransferList->Next( (CObject*&)obj) )
					{
						TransferObject = (CTransferObject*)obj;

						if ( (TransferObject->m_pTransfer->GetDone() == etsNONE) &&
						     (TransferObject->m_pTransfer->GetSrcDirection() == edUPLOAD) )
						{
							if ( (TransferObject->m_pTransfer->GetTransferrate() >= (TransferObject->m_pTransfer->GetRate()*3/4)) || ((ulonglong)rate < TransferObject->m_pTransfer->GetRate()) )
							{
								TransferObject->m_pTransfer->SetRate(tmprate);
							}
						}
					}
				}
			}
		}


		// send dl manager info
		SendDownloadManagerInfo(&DownloadManagerInfo);
		// send traffic info
		SendTrafficInfo();
	}

	m_pTransferList->UnLock();
}

/** */
void CDownloadManager::UpdateQueueList( time_t ttimeout )
{
	int i;
	CStringList * StringList = 0;
	CList<DCHubObject> hublist;
	DCHubObject * HubObject1,* HubObject2;
	CString hubname;

	m_pDownloadQueue->pQueue->Lock();

	while( m_pDownloadQueue->pQueue->Next( (CObject *&) StringList) )
	{
		DCTransferQueueObject * TransferObject = 0;

		while( StringList->Next( (CObject *&) TransferObject) )
		{
			if ( TransferObject->pTransferFileList.Count() == 0 )
			{
				// empty filelist and no connections
				if ( TransferObject->iConnections <= 0 )
				{
					SendFileInfo( TransferObject, 0, TRUE );
					StringList->Del(TransferObject->sHubName);
				}
				break;
			}
			else if ( (TransferObject->eState == etwsWAIT) && (m_eShutdownState == essNONE) )
			{
				// check response timeout
				if ( (ttimeout-TransferObject->tTimeout) >= CConfig::Instance()->GetTransferResponseTimeout() )
				{
					// set timeout
					TransferObject->eState = etwsIDLE;
					TransferObject->tTimeout = ttimeout;
					SendFileInfo( TransferObject );
				}
			}
			else if ( (m_eShutdownState == essNONE) &&
				( (TransferObject->eState == etwsIDLE) ||
				  (TransferObject->eState == etwsHUBOFFLINE) ||
				  (TransferObject->eState == etwsUSEROFFLINE) ||
				  (TransferObject->eState == etwsUSERBUSY) ||
				  (TransferObject->eState == etwsSENDERROR) ) )
			{

				// idle state ... check for timeout to reconnect
				if ( TransferObject->tTimeout == 0 )
				{
					// check for max download rate before new connection start
					if ( !((CConfig::Instance()->GetMaxDownloadRate() != 0)  &&
					     (CConfig::Instance()->GetMaxDownloadRate() < DownloadManagerInfo.rate_dl)) )
					{
						// reset hubname
						hubname = TransferObject->sHubName;

						// check if the user in a hub online
						if ( CServerManager::Instance()->IsUserOnline( TransferObject->sNick, "", &hublist ) == TRUE )
						{
							printf("user is online on:\n");

							HubObject1 = 0;
							while( (HubObject1=hublist.Next(HubObject1)) != 0 )
							{
								printf("'%s' '%s'\n",HubObject1->m_sHubName.Data(),HubObject1->m_sHubHost.Data());

								HubObject2 = 0;
								while( (HubObject2=TransferObject->pHubList.Next(HubObject2)) != 0 )
								{
									if ( HubObject1->m_sHubName == HubObject2->m_sHubName )
										break;
								}

								// add new entry
								if ( HubObject2 == 0 )
								{
									printf("NEW '%s' '%s'\n",HubObject1->m_sHubName.Data(),HubObject1->m_sHubHost.Data());
									HubObject2 = new DCHubObject();
									HubObject2->m_sHubName = HubObject1->m_sHubName;
									HubObject2->m_sHubHost = HubObject1->m_sHubHost;
									HubObject2->m_bActive = TRUE;

									TransferObject->pHubList.Add(HubObject2);
								}
								else // set entry to connect
								{
									if ( (hubname == "") && (HubObject2->m_bActive == TRUE) )
									{
										printf("USE '%s'\n",HubObject2->m_sHubName.Data());
										hubname = HubObject2->m_sHubName;
									}
								}
							}

							hublist.Clear();
						}

						// send connection request to the hub on timeout
						i = CServerManager::Instance()->SendConnectionRequest( TransferObject->sNick, hubname );

						switch (i)
						{
							case 0:
								TransferObject->eState = etwsWAIT;
								break;
							case -1:
								TransferObject->eState = etwsUSEROFFLINE;
								break;
							case -2:
							case -3:
								TransferObject->eState = etwsHUBOFFLINE;
								break;
							case -4:
								TransferObject->eState = etwsSENDERROR;
								break;
							default:
								break;
						}
					}

					TransferObject->tTimeout = ttimeout;
					SendFileInfo( TransferObject );
				}
				// check resend timeout
				else if ( (ttimeout-TransferObject->tTimeout) >= CConfig::Instance()->GetTransferResendTimeout() )
				{
					TransferObject->tTimeout = 0;
					SendFileInfo( TransferObject );
				}
			}
		}
	}

	m_pDownloadQueue->pQueue->UnLock();
}

/** thread callbackfunction */
int CDownloadManager::Callback( CObject *, CObject * )
{
	int i;
	time_t ttimeout;

	if ( m_eShutdownState == essSHUTDOWNREADY )
	{
		return 0;
	}

	ttimeout = time(0);

	if ( m_eShutdownState == essNONE )
	{
		// save queue
		i = CConfig::Instance()->GetDownloadQueueTime();

		if ( i > 0 )
		{
			i *= 60;

			if ( (ttimeout-tDownloadQueueTimeout) > i )
			{
				DLM_SaveQueue();
				tDownloadQueueTimeout = ttimeout;
			}
		}
	}

	// update the transfer list
	UpdateTransferList(ttimeout);

	// update the queue list
	if ( (ttimeout-tUpdateTransferTimeout) >= 1 )
	{
		UpdateQueueList(ttimeout);
	}

	// update timeout
	tUpdateTransferTimeout = ttimeout;

	return 0;
}

/** callback function */
int CDownloadManager::DM_ListenCallBack( CObject * /*Sender*/, CObject * Object )
{
	int handle = *((int*)Object);

	CDownloadManager::Instance()->m_pTransferWaitList->Lock();

	if ( CDownloadManager::Instance()->m_pTransferWaitList->Count() == 0 )
	{
		handle = -1;
	}

	CDownloadManager::Instance()->m_pTransferWaitList->UnLock();

	if ( handle == -1 )
	{
		return -1;
	}

	CTransferObject * TransferObject = new CTransferObject();
	TransferObject->m_pTransfer = new CTransfer(TRUE);

	TransferObject->m_pTransfer->SetTransferID( CDownloadManager::Instance()->GetNewTransferID() );
	TransferObject->m_pTransfer->SetNick( CConfig::Instance()->GetNick() );
	TransferObject->m_pTransfer->SetRate( CConfig::Instance()->GetMaxUploadRate() );
	TransferObject->m_pTransfer->SetCallBackFunction( new CCallback<CDownloadManager>( CDownloadManager::Instance(), &CDownloadManager::DM_TransferCallBack ) );

	if ( TransferObject->m_pTransfer->SetSocket(handle) == 0 )
	{
		CDownloadManager::Instance()->SendLogInfo("Incoming connection from '"+TransferObject->m_pTransfer->GetHost()+"'\n");

		CDownloadManager::Instance()->m_pTransferList->Lock();
		CDownloadManager::Instance()->m_pTransferList->Add( TransferObject->m_pTransfer->GetTransferID(), TransferObject );
		CDownloadManager::Instance()->m_pTransferList->UnLock();
	}
	else
	{
		delete TransferObject;
		close(handle);
	}

	return 0;
}

/** */
int CDownloadManager::DM_TransferCallBack( CObject * transfer, CObject * Object )
{
	CDownloadManager::Instance()->TransferCallBackThread.Lock();

	CDCMessage *DCMsg;
	CByteArray ba;
	CString s,r;
	ulonglong  len;
	CDir dir;
	CList<CString> SharedFolders;
	bool remove,bdirec;
	eShareBufferType stype;

	CTransfer * Transfer = (CTransfer*)transfer;

	DCMsg = (CDCMessage*)Object;

	switch ( DCMsg->m_eType )
	{
		case DC_MESSAGE_MYNICK:
		{
			// check with the nick for download or upload
			if ( CDownloadManager::Instance()->CheckWaitTransfer( Transfer ) == FALSE )
			{
				Transfer->SetSrcDirection(edUPLOAD);
			}
			else
			{
				Transfer->SetSrcDirection(edDOWNLOAD);
			}

			break;
		}

		case DC_MESSAGE_GET:
		{
			// remote want a file ...
			CMessageGet * msg = (CMessageGet*)Object;

			if ( (Transfer->GetSrcDirection() == edUPLOAD) && (Transfer->GetDstDirection() == edDOWNLOAD) )
			{
				// reset done flags
				Transfer->SetDone(etsNONE);

				// send a log info
				CDownloadManager::Instance()->SendLogInfo( "Upload: " + msg->sFilename, Transfer );

				// check for operator/special upload (only the filelist)
				if ( (Transfer->GetTransferType() == ettOPERATOR) &&
				     (msg->sFilename != CString(DC_USER_FILELIST_HE3)) &&
				     (msg->sFilename != CString(DC_USER_FILELIST_BZ)) )
				{
					// disconnect ...
					CDownloadManager::Instance()->SendLogInfo( "Operator Transfer not for the filelist", Transfer );
					Transfer->SendMaxedOut();
					Transfer->Disconnect(TRUE);
					break;
				}

				stype = esbtNONE;

				if ( msg->sFilename == CString(DC_USER_FILELIST_HE3) )
				{
					stype = esbtHE3;
				}
				else if ( msg->sFilename == CString(DC_USER_FILELIST_BZ) )
				{
					stype = esbtBZ;
				}

				// check special transfer file size
				if ( stype != esbtNONE )
				{
					// remote want the filelist
					if ( CConfig::Instance()->GetShareSize() == 0 )
					{
						Transfer->SendError("File not found.");

						CDownloadManager::Instance()->SendLogInfo( "Upload (sharesize is 0): " + msg->sFilename, Transfer );
					}
					else
					{
						if ( CConfig::Instance()->GetShareBuffer( stype, &ba ) == 0 )
						{
							// set send buffer
							Transfer->SetBuffer(&ba);
							// set local transfer medium to buffer
							Transfer->SetMedium(eltBUFFER);
							// start upload
							Transfer->StartUpload( msg->sFilename, CConfig::Instance()->GetShareBufferSize(stype), msg->lPos-1, 0, "" );
						}
						else
						{
							Transfer->SendError("File not found.");

							CDownloadManager::Instance()->SendLogInfo( "Upload (no sharebuffer): " + msg->sFilename, Transfer );
						}
					}
				}
				else // remote download from the share ...
				{
					// get the shared folder list
					if ( CConfig::Instance()->GetSharedFolders(&SharedFolders) == 0 )
					{
						Transfer->SendError("File not found.");

						CDownloadManager::Instance()->SendLogInfo( "Upload (no share folders): " + msg->sFilename, Transfer );
					}
					// search the wanted file in the share
					else if ( (s = dir.GetCorrectFilePath(msg->sFilename.Data(),&SharedFolders)) == "" )
					{
						Transfer->SendError("File not found.");

						CDownloadManager::Instance()->SendLogInfo( "Upload (File not found): " + msg->sFilename, Transfer );
					}
					else
					{
						// file found, check filesize
						len = dir.getFileSize(s,FALSE);

						if ( msg->lPos > len ) // if the wanted filepos > local length send an error
						{
							Transfer->SendError("File allready download.");
						}
						else if ( msg->lPos == 0 ) // we begin at 1 (see dc protocoll)
						{
							Transfer->SendError("Wrong file position.");
						}
						else // upload the wanted file ...
						{
							// special transfer only for files < 16k
							if ( (Transfer->GetTransferType() == ettSPECIAL) &&
							     (len > (1024*16)) )
							{
								// disconnect ...
								CDownloadManager::Instance()->SendLogInfo( "Special Transfer not for files > 16k", Transfer );
								Transfer->SendMaxedOut();
								Transfer->Disconnect(TRUE);
							}
							else
							{
								// set local transfer medium to buffer
								Transfer->SetMedium(eltFILE);
								// start upload
								Transfer->StartUpload( msg->sFilename, len, msg->lPos-1, msg->lSize, s );
							}
						}
					}
				}
			}
			else
			{
				CDownloadManager::Instance()->SendLogInfo( "Warning wrong mode", Transfer );
				Transfer->Disconnect(TRUE);
			}

			break;
		}

		case DC_MESSAGE_DIRECTION:
		{
			break;
		}

		case DC_MESSAGE_KEY:
		{
			// direction message ...
			//CMessageDirection * msg = (CMessageDirection*)Object;
			bdirec = FALSE;

			printf("DIRECTION: level: LOCAL: %d REMOTE: %d\n",Transfer->GetSrcLevel(), Transfer->GetDstLevel() );
			printf("DIRECTION: direc: LOCAL: %d REMOTE: %d\n",Transfer->GetSrcDirection(), Transfer->GetDstDirection() );

			// equal direction ...
			if ( Transfer->GetDstDirection() == Transfer->GetSrcDirection() )
			{
				// check the level ...
				if ( Transfer->GetDstLevel() < Transfer->GetSrcLevel() )
				{
					// now we must change the dst direction
					if ( Transfer->GetSrcDirection() == edDOWNLOAD )
					{
						Transfer->SetDstDirection(edUPLOAD);
						bdirec = TRUE;
					}
				}
				else if ( Transfer->GetDstLevel() > Transfer->GetSrcLevel() )
				{
					if ( Transfer->GetSrcDirection() == edDOWNLOAD )
					{
						Transfer->SetSrcDirection(edUPLOAD);
						bdirec = TRUE;
					}
				}
			}
			else
			{
				bdirec = TRUE;
			}

			if ( bdirec == TRUE )
			{
				if ( CDownloadManager::Instance()->SetDirection(Transfer) == FALSE )
				{
					bdirec = FALSE;
				}
			}

			if ( bdirec == FALSE )
			{
				CDownloadManager::Instance()->SendLogInfo( "Warning no more free slots", Transfer );
				Transfer->SetDstDirection(edNONE);
				Transfer->SendMaxedOut();
				Transfer->Disconnect(TRUE);
			}
			else
			{
				if ( Transfer->GetSrcDirection() == edDOWNLOAD )
				{
					printf("DIRECTION: download mode ...\n");

					// next file
					if ( CDownloadManager::Instance()->SetNextFile(Transfer) == FALSE )
					{
						printf("DIRECTION: download mode without destination file -> disconnecting!\n");
						Transfer->Disconnect(TRUE);
					}
				}
				else if ( Transfer->GetSrcDirection() == edUPLOAD )
				{
					printf("DIRECTION: we are in upload mode ...\n");
				}
			}

			break;
		}

		case DC_MESSAGE_CONNECTION_STATE:
		{
			// connection-state message, generated by the ctransfer-class
			remove = TRUE;
			CMessageConnectionState *msg = (CMessageConnectionState*)Object;

			if ( msg->eState == estCONNECTED )
			{
				remove = FALSE;
			}
			else if ( msg->eState == estDISCONNECTED )
			{
				CDownloadManager::Instance()->SendLogInfo( "Disconnected from "+Transfer->GetHost(), Transfer );
			}
			else if ( msg->eState == estSOCKETERROR )
			{
				remove = FALSE;
			}
			else if ( msg->eState == estCONNECTIONTIMEOUT )
			{
				remove = FALSE;
			}

			if ( (remove == TRUE) && (Transfer->GetSrcDirection() == edDOWNLOAD) )
			{
				if ( Transfer->GetMedium() == eltBUFFER )
					Transfer->SetStartPosition(0);
				CDownloadManager::Instance()->UpdateFileState(Transfer,etfsNONE);
			}

			if ( remove == TRUE )
			{
				Transfer->SetDone(etsREADY);
			}

			break;
		}

		case DC_MESSAGE_TRANSFER:
		{
			// transfer message, generated by the ctransfer-class
			CMessageTransfer *msg = (CMessageTransfer*)Object;

			// filetransfer done
			if ( (msg->lTransfered == msg->lLength) ||
			     ((Transfer->GetSupport().bChunk == TRUE) &&
			      (Transfer->GetStartPosition()+Transfer->GetChunkSize() == Transfer->GetCurrentPosition())) )
			{
				// download ready
				if ( Transfer->GetSrcDirection() == edDOWNLOAD )
				{
					CDownloadManager::Instance()->UpdateFileState(Transfer,etfsNONE);

					// show user file list
					if ( Transfer->GetDstFilename() == CString(DC_USER_FILELIST) )
					{
						CDownloadManager::Instance()->FileListDone(Transfer);
					}

					if ( (msg->lTransfered == Transfer->GetLength()) ||
					     ((Transfer->GetSupport().bChunk == TRUE) &&
					      (Transfer->GetStartPosition()+Transfer->GetChunkSize() == Transfer->GetCurrentPosition())) )
					{
						if ( CDownloadManager::Instance()->SetNextFile(Transfer) == FALSE )
						{
						}
					}
				}
				else // upload ready
				{
					// set timeout to save local slots
					if ( Transfer->GetDone() != etsIDLE )
					{
						Transfer->SetStartTime(time(0));
						Transfer->SetDone(etsIDLE);
					}
				}
			}
			else 
			{
				// change new chunk end only for file medium and if remote not support chunks
				if ( (Transfer->GetSrcDirection() == edDOWNLOAD) &&
				     (Transfer->GetMedium() == eltFILE) &&
				     (Transfer->GetSupport().bChunk == FALSE) )
				{
					ulonglong lstart,lend;

					if ( (msg->lLength-msg->lTransfered) < (CHUNK_SIZE/2) )
					{
						if ( CDownloadManager::Instance()->GetNewChunkEnd( Transfer->GetSrcFilename(), Transfer->GetStartPosition(), Transfer->GetEndPosition(), Transfer->GetCurrentPosition(), &lstart, &lend ) == TRUE )
						{
							Transfer->SetStartPosition(lstart);
							Transfer->SetEndPosition(lend);
						}
					}
				}
			}

			break;
		}

		case DC_MESSAGE_FILELENGTH:
		{
			// remote send a filelength
			CMessageFileLength * msg = (CMessageFileLength*)Object;

			if ( msg->lFileLength == 0 )
			{
				CDownloadManager::Instance()->SendLogInfo( "Warning: Filelength is NULL.", Transfer );

				if ( Transfer->GetSrcDirection() == edDOWNLOAD )
				{
					CDownloadManager::Instance()->UpdateFileState(Transfer,etfsERROR);
				}

				Transfer->Disconnect(TRUE);
			}
			else
			{
				CDownloadManager::Instance()->SendTransferInfo(Transfer);
			}

			break;
		}

		case DC_MESSAGE_MAXEDOUT:
		{
			// no free slots on the remote side
			CDownloadManager::Instance()->SendLogInfo( "Busy", Transfer );
			Transfer->Disconnect(TRUE);
			break;
		}

		case DC_MESSAGE_GETLISTLEN:
		{
			// remote want the listlen
			Transfer->SendListLen(CConfig::Instance()->GetShareBufferSize(esbtHE3));
			break;
		}

		case DC_MESSAGE_ERROR:
		{
			// remote send a error message
			CMessageError * msg = (CMessageError*)Object;

			CDownloadManager::Instance()->SendLogInfo( "Error: " + msg->sError, Transfer );

			if ( Transfer->GetSrcDirection() == edDOWNLOAD )
			{
				CDownloadManager::Instance()->UpdateFileState(Transfer,etfsERROR);

				// set next file for download
				if ( CDownloadManager::Instance()->SetNextFile(Transfer) == FALSE )
				{
				}
			}
			break;
		}

		default:
		{
			printf("dctransfer unknown message: %d\n",DCMsg->m_eType);
			break;
		}
	}

	delete Object;

	CDownloadManager::Instance()->TransferCallBackThread.UnLock();

	return 0;
}
