/***************************************************************************
                           cfilemanager.cpp  -  description
                             -------------------
    begin                : Sat Aug 03 2002
    copyright            : (C) 2002-2003 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef WIN32
#include <unistd.h>
#include <stdlib.h>
#endif

#include <dclib/dcos.h>
//#include <dclib/cmd4.h>
#include <dclib/cmd5.h>
#include <dclib/cconfig.h>
#include <dclib/cstring.h>
#include <dclib/cstringlist.h>
#include <dclib/cdir.h>
#include <dclib/cdownloadmanager.h>

#include "cfilemanager.h"

class CFileType : public CObject {
public:
	/** */
	CFileType() {};
	/** */
	virtual ~CFileType() {};

	/** */
	eFileTypes m_eFileType;
};

/** filetype list */
typedef struct filetype {
	const char* sExt;
	eFileTypes eFileType;
} filetype;

struct filetype FileType[] = {
	// MP3 2
	{ "AIFF", eftMP3 },
	{ "AU",   eftMP3 },
	{ "MID",  eftMP3 },
	{ "MP2",  eftMP3 },
	{ "MP3",  eftMP3 },
	{ "OGG",  eftMP3 },
	{ "RA",   eftMP3 },
	{ "WAV",  eftMP3 },

	// ARCHIVE 3
	{ "ACE", eftARCHIVE },
	{ "BZ2", eftARCHIVE },
	{ "CAB", eftARCHIVE },
	{ "EX_", eftARCHIVE },
	{ "GZ",  eftARCHIVE },
	{ "LZH", eftARCHIVE },
	{ "RAR", eftARCHIVE },
	{ "RPM", eftARCHIVE },
	{ "TAR", eftARCHIVE },
	{ "TGZ", eftARCHIVE },
	{ "ZIP", eftARCHIVE },
	{ "ZOO", eftARCHIVE },
	{ "Z",   eftARCHIVE },

	// DOCUMENT 4
	{ "CFG",   eftDOCUMENT },
	{ "CPP",   eftDOCUMENT },
	{ "DIZ",   eftDOCUMENT },
	{ "DOC",   eftDOCUMENT },
	{ "H",     eftDOCUMENT },
	{ "HLP",   eftDOCUMENT },
	{ "HTM",   eftDOCUMENT },
	{ "HTML",  eftDOCUMENT },
	{ "INI",   eftDOCUMENT },
	{ "INF",   eftDOCUMENT },
	{ "LOG",   eftDOCUMENT },
	{ "PHP",   eftDOCUMENT },
	{ "PS",    eftDOCUMENT },
	{ "PDF",   eftDOCUMENT },
	{ "SHTML", eftDOCUMENT },
	{ "TXT",   eftDOCUMENT },
	{ "RFT",   eftDOCUMENT },
	{ "XML",   eftDOCUMENT },

	// APPL 5
	{ "BAT",  eftAPPLICATION },
	{ "CGI",  eftAPPLICATION },
	{ "COM",  eftAPPLICATION },
	{ "DLL",  eftAPPLICATION },
	{ "EXE",  eftAPPLICATION },
	{ "HQX",  eftAPPLICATION },
	{ "JS",   eftAPPLICATION },
	{ "SO",   eftAPPLICATION },
	{ "SYS",  eftAPPLICATION },
	{ "VXD",  eftAPPLICATION },

	// PICTURE 6
	{ "3DS",  eftPICTURE },
	{ "A11",  eftPICTURE },
	{ "ACB",  eftPICTURE },
	{ "ADC",  eftPICTURE },
	{ "ADI",  eftPICTURE },
	{ "AFI",  eftPICTURE },
	{ "AI",   eftPICTURE },
	{ "AIS",  eftPICTURE },
	{ "ANS",  eftPICTURE },
	{ "ART",  eftPICTURE },
	{ "B8",   eftPICTURE },
	{ "BMP",  eftPICTURE },
	{ "CBM",  eftPICTURE },
	{ "GIF",  eftPICTURE },
	{ "ICO",  eftPICTURE },
	{ "JPEG", eftPICTURE },
	{ "JPG",  eftPICTURE },
	{ "PIC",  eftPICTURE },
	{ "PNG",  eftPICTURE },
	{ "PCX",  eftPICTURE },
	{ "TGA",  eftPICTURE },
	{ "TIF",  eftPICTURE },

	// VIDEO 7
	{ "AVI",   eftVIDEO },
	{ "ASF",   eftVIDEO },
	{ "ASX",   eftVIDEO },
	{ "DAT",   eftVIDEO },
	{ "MOV",   eftVIDEO },
	{ "MOVIE", eftVIDEO },
	{ "MPG",   eftVIDEO },
	{ "MP4",   eftVIDEO },
	{ "MPEG",  eftVIDEO },
	{ "OGM",   eftVIDEO },
	{ "QT",    eftVIDEO },
	{ "RM",    eftVIDEO },
	{ "VIV",   eftVIDEO },
	{ "WMV",   eftVIDEO },
	{ "DIVX",  eftVIDEO },

	// END
	{ 0, eftALL }
};

/** */
CFileManager::CFileManager()
{
	pFileList         = new CThreadStringList();
	pFileBaseList     = new CThreadStringList();
	pQueryStringIndex = new CStringList(5);
	m_pFileTypeList   = new CStringList();
	m_pFileNameList   = 0;
	m_pFileBaseObject = 0;

	bCreateIndex     = FALSE;
	bCreateHashList  = FALSE;

	FileManagerInfo = new CFileManagerInfo();
	FileManagerInfo->m_bCreateShareList   = FALSE;
	FileManagerInfo->m_bCreateSearchIndex = FALSE;
	FileManagerInfo->m_nShareListPercent  = 100;

	InitFileTypeList();

	CConfig::Instance()->LoadDCShare( pFileBaseList );

	SetInstance(this);
}

/** */
CFileManager::~CFileManager()
{
	Stop();

	Lock();

	if ( pFileList )
	{
		delete pFileList;
		pFileList = 0;
	}

	if ( pFileBaseList )
	{
		delete pFileBaseList;
		pFileBaseList = 0;
	}

	if ( pQueryStringIndex )
	{
		delete pQueryStringIndex;
		pQueryStringIndex = 0;
	}

	if ( m_pFileNameList )
	{
		delete m_pFileNameList;
		m_pFileNameList = 0;
	}

	if ( FileManagerInfo )
	{
		delete FileManagerInfo;
		FileManagerInfo = 0;
	}

	delete m_pFileTypeList;

	UnLock();
}

/** */
void CFileManager::InitFileTypeList()
{
	int i;
	CFileType * filetype;

	for(i=0;FileType[i].sExt != 0;i++)
	{
		filetype = new CFileType();
		filetype->m_eFileType = FileType[i].eFileType;

		m_pFileTypeList->Add( FileType[i].sExt, filetype );
	}
}

/** return the filetype */
eFileTypes CFileManager::GetFileType( CString file )
{
	CString ext;
	eFileTypes res=eftUNKNOWN;
	CFileType * filetype;

	// set the filetype
	ext = CDir::Extension(file);

	if ( ext != "" )
	{
		if ( m_pFileTypeList->Get( ext.ToUpper(), (CObject*&)filetype ) == 0 )
		{
			res = filetype->m_eFileType;
		}
#if 0
		if ( res == eftUNKNOWN )
			printf("unknown extension: %s\n",ext.Data());
#endif
	}

	return res;
}

/** */
CStringList * CFileManager::Search( CString s )
{
	CString *st,st1;
	CStringList *sl, *sl1 = 0;

	if ( !FileManagerInfo )
	{
		return 0;
	}

	if ( (FileManagerInfo->m_bCreateShareList == TRUE) ||
	     (FileManagerInfo->m_bCreateSearchIndex == TRUE) )
	{
		return 0;
	}

	Lock();

	if ( pQueryStringIndex )
	{
		if ( pQueryStringIndex->Get( s, (CObject*&)sl ) == 0 )
		{
			sl1 = new CStringList();

			st = 0;
			while( sl->Next(st1,(CObject*&)st) != 0 )
			{
				sl1->Add(st1,new CString(*st));
			}
		}
	}

	UnLock();

	return sl1;
}

/** */
CFileBaseObject * CFileManager::GetFileBaseObject( CString id )
{
	CFileBaseObject * FileBaseObject = 0, * FileBaseObject1 = 0;

	if ( !FileManagerInfo )
	{
		return 0;
	}

	if ( (FileManagerInfo->m_bCreateShareList == TRUE) ||
	     (FileManagerInfo->m_bCreateSearchIndex == TRUE) )
	{
		return 0;
	}

	Lock();

	if ( pFileBaseList->Get( id, (CObject*&)FileBaseObject ) == 0 )
	{
		FileBaseObject1 = new CFileBaseObject(FileBaseObject);
	}

	UnLock();

	return FileBaseObject1;
}

/** */
void CFileManager::SaveIndex()
{
	if ( !FileManagerInfo )
	{
		return;
	}

	if ( FileManagerInfo->m_bCreateSearchIndex == FALSE )
	{
		CConfig::Instance()->SaveDCShare( pFileBaseList );
	}
}

/** */
void CFileManager::StartIndex()
{
	pFileList->Clear();
	pFileBaseList->Clear();
}

/** */
void CFileManager::StopIndex()
{
	SaveIndex();
}

/** */
void CFileManager::AddIndex( CFileInfo *fileinfo, CString path, CString /*rootpath*/ )
{
	CFileBaseObject * FileBaseObject;

	if ( !pFileBaseList )
	{
		return;
	}

	FileBaseObject = new CFileBaseObject();

	FileBaseObject->ID        = pFileBaseList->Count()+1;
	FileBaseObject->lSize     = fileinfo->size;
	FileBaseObject->sFile     = fileinfo->name;
	FileBaseObject->sPath     = path;
	FileBaseObject->eFileType = GetFileType(fileinfo->name);
//	FileBaseObject->sRootPath = rootpath;
//	FileBaseObject->st_a_time = fileinfo->st_a_time;
//	FileBaseObject->st_c_time = fileinfo->st_c_time;
//	FileBaseObject->st_m_time = fileinfo->st_m_time;

	// add to the base list
	pFileBaseList->Add( CString().setNum(FileBaseObject->ID), FileBaseObject );
}

/** */
bool CFileManager::CreateShareList()
{
	bool err = FALSE;

	Lock();

	if ( (FileManagerInfo->m_bCreateSearchIndex == FALSE) &&
	     (FileManagerInfo->m_bCreateShareList == FALSE) )
	{
		SharedFolders.Clear();

		pShareFolder     = 0;
		iShareSize       = 0;
		sShareBuffer     = "";

		pFileList->Clear();
		pQueryStringIndex->Clear();

		if ( CConfig::Instance()->GetSharedFolders(&SharedFolders) != 0 )
		{
			if ( m_pFileNameList )
				delete m_pFileNameList;
			m_pFileNameList  = new CStringList();

			FileManagerInfo->m_nShareListPercent = 0;
			FileManagerInfo->m_bCreateShareList  = TRUE;

			if ( CDownloadManager::Instance() )
				CDownloadManager::Instance()->SendFileManagerInfo(FileManagerInfo);

			err = TRUE;
		}
		else
		{
			CConfig::Instance()->SetShareBuffer( "", 0 );
		}
	}

	if ( err == TRUE )
	{	
		StartIndex();

		Start();
	}

	UnLock();

	return err;
}

/** */
void CFileManager::CreateSearchIndex()
{
	if ( (FileManagerInfo->m_bCreateSearchIndex == FALSE) &&
	     (FileManagerInfo->m_bCreateShareList == FALSE) )
	{
		pFileList->Clear();
		pQueryStringIndex->Clear();

		m_pFileBaseObject = 0;
		m_nSearchIndex    = 0;

		FileManagerInfo->m_bCreateSearchIndex = TRUE;

		Start();
	}
}

/** */
bool CFileManager::CreateHashList()
{
	return FALSE;
}

/** */
void CFileManager::ThreadCreateSearchIndex()
{
	int i,k;
	CString s,s1,id;
	CStringList *sl;
	CString * ps;

	for(k=0;k<100;k++)
	{
	if ( pFileBaseList->Next( (CObject*&)m_pFileBaseObject) )
	{
		m_nSearchIndex++;

		id = CString().setNum(m_pFileBaseObject->ID);
		s  = m_pFileBaseObject->sPath + DIRSEPARATOR + m_pFileBaseObject->sFile;

		// create the search index
		pFileList->Add(s,new CString(id));

		// use the filename for the index
		s = m_pFileBaseObject->sFile;
		s = s.ToUpper();

		// add complete filename to the list
		if ( pQueryStringIndex->Get( s, (CObject*&)sl ) != 0 )
		{
			sl = new CStringList();
			pQueryStringIndex->Add( s, sl );
		}

		sl->Add( id, new CString(id) );

		// convert special chars
		s = s.Replace(' ',"/");
		s = s.Replace('.',"/");
		s = s.Replace('-',"/");
		s = s.Replace('_',"/");
		s = s.Replace('(',"/");
		s = s.Replace(')',"/");
		s = s.Replace('!',"/");

		s = CDir().ConvertSeparators(s);

		while( (i=s.Find(DIRSEPARATOR)) != -1 )
		{
			s1 = s.Left(i);
			s  = s.Mid(i+1,s.Length()-i-1);

			i++;

			// max length is 3
			if ( s1.Length() < 3 )
				continue;

			if ( pQueryStringIndex->Get( s1, (CObject*&)sl ) != 0 )
			{
				sl = new CStringList();
				pQueryStringIndex->Add( s1, sl );
			}

			if ( sl->Get( id, (CObject*&)ps ) != 0 )
			{
				sl->Add( id, new CString(id) );
			}
		}
	}
	else
	{
		printf("ready create index\n");

		FileManagerInfo->m_bCreateSearchIndex = FALSE;

		Stop();

		return;
	}
	}
}

/** */
void CFileManager::ThreadCreateShareList()
{
	CString s,s1;
	CDir dir;

	if ( (pShareFolder=SharedFolders.Next(pShareFolder)) != 0 )
	{
		// directory
		s = (*pShareFolder);

		// is valid ?
		if ( dir.cd(s.Data()) == TRUE )
		{
			s  = dir.Path();
			s1 = dir.DirName();

			if ( s1 != "" )
			{
				s = s.Left( s.Length()-s1.Length() );
			}

			ThreadCreateShareList( 0, s, s1, "", 100/SharedFolders.Count() );
		}
		else
		{
			printf("Can't change dir: '%s'\n",s.Data());
		}
	}
	else
	{
		FileManagerInfo->m_nShareListPercent = 100;
		printf("ready create sharelist\n");

		CConfig::Instance()->SetShareBuffer( sShareBuffer, iShareSize );

		if ( m_pFileNameList )
		{
			delete m_pFileNameList;
			m_pFileNameList = 0;
		}

		FileManagerInfo->m_bCreateShareList = FALSE;

		if ( CDownloadManager::Instance() )
			CDownloadManager::Instance()->SendFileManagerInfo(FileManagerInfo);

		StopIndex();

		CreateSearchIndex();

		if ( CDownloadManager::Instance() )
			CDownloadManager::Instance()->SendFileManagerInfo(FileManagerInfo);
	}
}

/** */
void CFileManager::ThreadCreateShareList( int depth, CString base, CString curr, CString relpath, double percent )
{
	CDir dir;
	CList<CFileInfo> list;
	CString d,s,*ts=0;
	CString path;
	CString rpath;
	CFileInfo *fileinfo;
	int x,z;

	if ( relpath == "" )
		rpath = curr;
	else
		rpath = relpath+DIRSEPARATOR+curr;

	if (base=="")
		path = curr;
	else
		path = base+DIRSEPARATOR+curr;

	dir.SetPath(path);

	d = "";
	z = depth;

//	printf("%s\n",curr.Data());

	// create depth string
	if ( curr != "" )
	{
		for(x=depth;x>0;x--) d+="\x9";
		z++;
		sShareBuffer += d + curr + "\xd\xa";
		d += "\x9";
	}

	// get all files
	if ( dir.ReadEntrys( CDir::Files, &list ) )
	{
		// add files
		fileinfo = 0;
		while( (fileinfo=list.Next(fileinfo)) != 0 )
		{
			if ( fileinfo->name != "" )
			{
				if ( fileinfo->m_bSymlink )
					s = dir.ReadLink(path+DIRSEPARATOR+fileinfo->name);
				else
					s = dir.SimplePath(path+DIRSEPARATOR+fileinfo->name);

				if ( s != "" )
				{
					if ( m_pFileNameList->Get(s,(CObject*&)ts) != 0 )
					{
						m_pFileNameList->Add(s,new CString());
						iShareSize   += fileinfo->size;
						sShareBuffer += d + fileinfo->name + "|" + CString().setNum(fileinfo->size) + "\xd\xa";
					}
				}

				// add file to the file manager ...
				AddIndex( fileinfo, rpath, path );
			}
		}
	}

	// get all dirs
	if ( dir.ReadEntrys( CDir::Dirs, &list ) )
	{
		if ( list.Count() > 0 )
			percent /= list.Count();
		else
			FileManagerInfo->m_nShareListPercent += percent;

		// add dir
		fileinfo = 0;
		while( (fileinfo=list.Next(fileinfo)) != 0 )
		{
			if ( (fileinfo->name != ".") && (fileinfo->name != "..") )
			{
				if ( fileinfo->m_bSymlink )
					s = dir.ReadLink(path+DIRSEPARATOR+fileinfo->name);
				else
					s = dir.SimplePath(path+DIRSEPARATOR+fileinfo->name);

				if ( s != "" )
				{
					if ( m_pFileNameList->Get(s,(CObject*&)ts) != 0 )
					{
						m_pFileNameList->Add(s,new CString());
						ThreadCreateShareList(z,path,fileinfo->name,rpath,percent);
					}
					else
						FileManagerInfo->m_nShareListPercent += percent;

				}
			}
			else
				FileManagerInfo->m_nShareListPercent += percent;

			if ( CDownloadManager::Instance() )
				CDownloadManager::Instance()->SendFileManagerInfo(FileManagerInfo);
		}
	}
}

/** */
ulonglong CFileManager::GetShareSize()
{
	CString s,s1;
	CDir dir;
	ulonglong iSSize;

	iSSize = 0;
	pShareFolder=0;

	if ( CConfig::Instance()->GetSharedFolders(&SharedFolders) == 0 )
	{
		printf("No share folderfound !");
		return 0;
	}

	while ( (pShareFolder=SharedFolders.Next(pShareFolder)) != 0 )
	{
		// directory
		s = (*pShareFolder);

		// is valid ?
		if ( dir.cd(s.Data()) == TRUE )
		{
			s  = dir.Path();
			s1 = dir.DirName();

			if ( s1 != "" )
			{
				s = s.Left( s.Length()-s1.Length() );
			}

			iSSize += GetShareSize( 0, s, s1, "" );
		}
		else
		{
			printf("Can't change to dir: '%s'\n",s.Data());
		}
	}

	return iSSize;
}


/** */
ulonglong CFileManager::GetShareSize( int depth, CString base, CString curr, CString relpath )
{
	ulonglong lShareSize;
	CDir dir;
	CList<CFileInfo> list;
	CString d,s;
	CString path;
	CString rpath;
	CFileInfo *fileinfo;
	int x,z;

	lShareSize = 0;

	if ( relpath == "" )
		rpath = curr;
	else
		rpath = relpath+DIRSEPARATOR+curr;

	if (base=="")
		path = curr;
	else
		path = base+DIRSEPARATOR+curr;

	dir.SetPath(path);

	d = "";
	z = depth;

	// create depth string
	if ( curr != "" )
	{
		for(x=depth;x>0;x--) d+="\x9";
		z++;
		d += "\x9";
	}

	// get all files
	if ( dir.ReadEntrys( CDir::Files, &list ) )
	{
		fileinfo = 0;
		while( (fileinfo=list.Next(fileinfo)) != 0 )
		{
			if ( fileinfo->name != "" )
			{
				if ( fileinfo->m_bSymlink )
					s = dir.ReadLink(path+DIRSEPARATOR+fileinfo->name);
				else
					s = dir.SimplePath(path+DIRSEPARATOR+fileinfo->name);

				if ( s != "" )
				{
					lShareSize += fileinfo->size;
				}
			}
		}
	}

	// get all dirs
	if ( dir.ReadEntrys( CDir::Dirs, &list ) )
	{
		fileinfo = 0;
		while( (fileinfo=list.Next(fileinfo)) != 0 )
		{
			if ( (fileinfo->name != ".") && (fileinfo->name != "..") )
			{
				if ( fileinfo->m_bSymlink )
					s = dir.ReadLink(path+DIRSEPARATOR+fileinfo->name);
				else
					s = dir.SimplePath(path+DIRSEPARATOR+fileinfo->name);

				if ( s != "" )
				{
					lShareSize += GetShareSize(z,path,fileinfo->name,rpath);
				}
			}
		}
	}
	return lShareSize;
}

/** */
void CFileManager::ThreadCreateHashList()
{
/*
	CFileBaseObject * FileBaseObject = 0;
	CString s;
	FILE * file;
	CMD5 * md5;
	MD4 * md4;
	int hashes;
	unsigned char * buffer, *phashes;
	int len;
	int i,j;

	while ( pFileBaseList->Next( (CObject*&)FileBaseObject) )
	{
		if ( (FileBaseObject->sMD5Hash == "") &&
		     (FileBaseObject->lSize > 0) )
		{
			// create the hash
			s = FileBaseObject->sRootPath + DIRSEPARATOR + FileBaseObject->sFile;

			printf("CREATE HASH: %s\n",s.Data());

			if ( (file=fopen(s.Data(),"rb")) != 0 )
			{
				hashes  = (FileBaseObject->lSize / EDONKEY_HASH_BLOCK_SIZE) + 1;
				phashes = (unsigned char *) malloc(hashes*16);
				buffer  = (unsigned char *) malloc(EDONKEY_HASH_BLOCK_SIZE/10);

//				printf("create %d md4 chunk hashes %llu\n",hashes,FileBaseObject->lSize);
				md5 = new CMD5();

				i = j = 0;
				md4 = 0;

				while ( !Stopped() )
				{
					len = fread(buffer, 1, EDONKEY_HASH_BLOCK_SIZE/10, file);

//					printf("%d %d : %d\n",i,j,len);

					if ( md4 == 0 )
					{
						md4 = new MD4();
					}

					if ( len > 0 )
					{
						md4->update(buffer, len);
						md5->update(buffer, len);
					}

					j++;

					if ( (j == 10) || (len == 0) )
					{
						md4->finalize();
						md4->raw_digest( phashes + (i*16) );

						delete md4;
						md4 = 0;
						i++;
						j = 0;
					}

					if ( (len == 0) || (i == hashes) )
					{
						break;
					}

					NanoSleep(25);
				}

//				printf("ready %d %d\n",i,hashes);

				if ( !Stopped() )
				{
					if ( hashes > 1 )
					{
						md4 = new MD4();
						md4->update(phashes, hashes*16);
						md4->finalize();
						FileBaseObject->sMD4Hash = md4->hex_digest();

						for(i=0;i<hashes;i++)
						{
							s = md4->hex_digest( phashes + (i*16) );
							FileBaseObject->pMD4ChunkHashList.Add( new CString(s) );
						}

						delete md4;
					}
					else
					{
						s = md4->hex_digest( phashes );
						FileBaseObject->sMD4Hash = s;
						FileBaseObject->pMD4ChunkHashList.Add( new CString(s) );
					}

					md5->finalize();

					FileBaseObject->sMD5Hash = md5->hex_digest();
				}

				delete md5;

				free(buffer);
				free(phashes);

				fclose(file);
				break;
			}
			else
			{
				printf("ERROR: %s\n",FileBaseObject->sFile.Data());
			}
		}
	}
*/
/*
	if ( (FileBaseObject == 0) && (!Stopped()) )
	{
		bCreateIndex = FALSE;
		Stop();
	}
*/
}

/** */
void CFileManager::Thread( CObject * )
{
	int wait = 50;

	Lock();

	if ( FileManagerInfo->m_bCreateShareList == TRUE )
	{
		ThreadCreateShareList();
	}
	else if ( FileManagerInfo->m_bCreateSearchIndex == TRUE )
	{
		ThreadCreateSearchIndex();

		wait = 0;
	}

	UnLock();

	NanoSleep(wait);
}
