/***************************************************************************
                          chubsearch.cpp  -  description
                             -------------------
    begin                : Fri Mar 15 2002
    copyright            : (C) 2002-2003 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>

#ifndef WIN32
#include <unistd.h>
#else
#include <io.h>
#endif

#include <dclib/dcos.h>
#include <dclib/dcobject.h>
#include <dclib/cconfig.h>
#include <dclib/cclient.h>
#include <dclib/cmanager.h>
#include <dclib/cmessagehandler.h>
#include <dclib/cservermanager.h>

#include "chubsearch.h"

CHubSearch * _pHubSearch;

CHubSearch::CHubSearch()
{
	sBuffer        = "";
	pHubServerList = 0;
	m_pSearchList  = 0;
	ehSearchType   = ehstNONE;

	pHubSearchClientList = new CList<CObject>();

	_pHubSearch = this;

	m_pCallback = new CCallback<CHubSearch>( this, &CHubSearch::DC_ManagerCallBack );
	CManager::Instance()->Add( m_pCallback );
}

CHubSearch::~CHubSearch()
{
	CManager::Instance()->Remove( m_pCallback );
	if ( m_pCallback )
	{
		delete m_pCallback;
		m_pCallback = 0;
	}

	// clear the client list
	HubSearchClientListThread.Lock();

	sHubSearchClient * HubSearchClient;

	if ( pHubSearchClientList != 0 )
	{
		while ( ( (HubSearchClient = (sHubSearchClient *)pHubSearchClientList->Next(0)) != NULL) )
		{
			HubSearchClient->Client->SetCallBackFunction(0);
			HubSearchClient->Client->Disconnect(TRUE);
			delete HubSearchClient->Client;
			HubSearchClient->Client = 0;
			pHubSearchClientList->Remove(HubSearchClient);
			delete HubSearchClient;
		}

		delete pHubSearchClientList;
		pHubSearchClientList = 0;
	}

	HubSearchClientListThread.UnLock();

	// clear hubserverlist
	if ( pHubServerList != 0 )
	{
		delete pHubServerList;
		pHubServerList = 0;
	}
}

/** */
void CHubSearch::ConnectionState( eConnectionState state )
{
	CMessageConnectionState *Object;

	Object = new CMessageConnectionState();

	Object->m_eType = DC_MESSAGE_CONNECTION_STATE;

	Object->eState   = state;
	Object->sMessage = GetSocketError();

	if ( DC_CallBack( Object ) == -1 )
	{
		printf("CallBack failed (state)...\n");
		delete Object;
	}
}

/** */
int CHubSearch::HandleMessage( char * c, int )
{
	int p=0;
	eDCMessage type;
	CObject * Object;

	CMessageHandler MessageHandler;

	CString s;

	s = c;

	while( (type=MessageHandler.Parse(&s,p,Object=0)) != DC_MESSAGE_PARSE_ERROR )
	{
		switch (type)
		{
			case DC_MESSAGE_SEARCHRESULT:
			{
				break;
			}

			default:
			{
				// remove message
				delete Object;
				Object = 0;
			}
		}

		if (Object)
		{
			((CDCMessage*)Object)->m_eType = type;

			if ( DC_CallBack( Object ) == -1 )
			{
				printf("CallBack failed ...\n");
				delete Object;
			}
		}
	}

	return 0;
}

/** */
void CHubSearch::DataAvailable( const char * buffer, int len )
{
	sBuffer = buffer;

	// add traffic control
	CSocket::m_Traffic.AddTraffic(ettCONTROLRX,len);

	// add '|' for parser
	sBuffer += '|';

	HandleMessage( (char*)sBuffer.Data(), sBuffer.Length() );
}

/** */
int CHubSearch::StartSearch( eHubSearchHubs hubs, CList<CObject> * searchlist, CString s )
{
	CObject * object;
	sHubSearchClient * HubSearchClient;

	ehSearchType = ehstNONE;

	// reset current server
	sCurrentServer = 0;

	// set start time
	tStartTime = time(0);

	// clear client list
	HubSearchClientListThread.Lock();

	while ( ( (HubSearchClient = (sHubSearchClient *)pHubSearchClientList->Next(0)) != NULL) )
	{
		pHubSearchClientList->Remove(HubSearchClient);

		HubSearchClient->Client->SetCallBackFunction(0);
		HubSearchClient->Client->Disconnect();
		delete HubSearchClient->Client;
		HubSearchClient->Client = 0;
		delete HubSearchClient;
	}

	HubSearchClientListThread.UnLock();

	// reset values
	iCurrentHubServer = 0;

	if ( pHubServerList != 0 )
	{
		delete pHubServerList;
		pHubServerList = 0;
	}

	m_pSearchList = searchlist;

	// check if a user search in the list
	m_bHandleUserList = FALSE;
	object = 0;

	while ( (object=m_pSearchList->Next(object)) != 0 )
	{
		if ( ((CDCMessage*)object)->m_eType == DC_MESSAGE_SEARCH_USER )
		{
			m_bHandleUserList = TRUE;
			break;
		}
	}
	
	if ( hubs == ehshCONNECTEDALL )
	{
		ehSearchType = ehstLOCAL;

		if ( SendSearch("") == 0 )
		{
			ehSearchType = ehstNONE;

			return -1;
		}

		ehSearchType = ehstWAITTIMEOUT;
	}
	else if ( hubs == ehshCONNECTEDSINGLE )
	{
		ehSearchType = ehstLOCAL;

		if ( SendSearch(s) == 0 )
		{
			ehSearchType = ehstNONE;

			return -1;
		}

		ehSearchType = ehstWAITTIMEOUT;
	}
	else
	{
		if ( hubs == ehshPUBLIC )
		{
			// get public hub server list
			pHubServerList = CServerManager::Instance()->GetPublicHubServerList();
		}
		else if ( hubs == ehshBOOKMARK )
		{
			// get bookmark hub server list
			pHubServerList = CServerManager::Instance()->GetBookmarkHubServerList();
		}

		if ( (pHubServerList == 0) || (pHubServerList->Count() <= 0) )
		{
			return -1;
		}

		ehSearchType = ehstGLOBAL;
	}

	Start();

	return 0;
}

/** */
void CHubSearch::StopSearch()
{
	ehSearchType = ehstSTOP;

	Disconnect();

//	Stop();
}

/** */
void CHubSearch::SendDebug( CString s )
{
	CMessageLog * MessageLog = new CMessageLog();

	MessageLog->sMessage = s;
	MessageLog->m_eType  = DC_MESSAGE_LOG;

	if ( DC_CallBack( MessageLog ) == -1 )
	{
		printf("CallBack failed (state)...\n");
		delete MessageLog;
	}
}

/** */
void CHubSearch::SendSearchResult( CObject * object )
{
	if ( DC_CallBack( object ) == -1 )
	{
		printf("CallBack failed (state)...\n");
		delete object;
	}
}

/** */
void CHubSearch::UpdateClients()
{
	if ( pHubSearchClientList != 0 )
	{
		sHubSearchClient * HubSearchClient = 0;

		while ( ( (HubSearchClient = (sHubSearchClient *)pHubSearchClientList->Next(HubSearchClient)) != NULL) )
		{
			HubSearchClient->Client->Thread(0);
		}
	}
}

/** */
int CHubSearch::DC_ManagerCallBack( CObject *, CObject * )
{
	UpdateClients();

	if ( (_pHubSearch->ehSearchType == ehstGLOBAL) ||
	     (_pHubSearch->ehSearchType == ehstWAITTIMEOUT) ||
	     (_pHubSearch->ehSearchType == ehstSTOP) )
	{
		_pHubSearch->CheckClient();

		if ( _pHubSearch->ehSearchType == ehstGLOBAL )
		{
			_pHubSearch->NewClient();
		}

		if ( _pHubSearch->ehSearchType == ehstSTOP )
		{
			_pHubSearch->HubSearchClientListThread.Lock();

			if ( _pHubSearch->pHubSearchClientList->Count() == 0 )
			{
				_pHubSearch->ehSearchType = ehstNONE;

				Stop();
			}

			_pHubSearch->HubSearchClientListThread.UnLock();
		}
	}

	return 0;
}

/** */
void CHubSearch::CheckClient()
{
	sHubSearchClient * HubSearchClient,* OldHubSearchClient;

	HubSearchClientListThread.Lock();

	if ( pHubSearchClientList != 0 )
	{
		HubSearchClient = OldHubSearchClient = 0;

		// check all hubs in the list
		while ( ( (HubSearchClient = (sHubSearchClient *)pHubSearchClientList->Next(HubSearchClient)) != NULL) )
		{
			if ( (ehSearchType == ehstSTOP) && (HubSearchClient->bRemove != TRUE) )
			{
				// set flag to remove all clients
				HubSearchClient->Client->SetCallBackFunction(0);
				HubSearchClient->bRemove = TRUE;

				OldHubSearchClient = HubSearchClient;
			}
			// remove the client or if timeout
			else if ( (HubSearchClient->bRemove == TRUE) || ((time(0)-HubSearchClient->timeHubTimeout) >= 300) )
			{
				// check remove timeout
				if ( (HubSearchClient->bRemove == TRUE) &&
				     ((time(0)-HubSearchClient->timeHubTimeout) < 6) &&
				     (ehSearchType != ehstSTOP) )
				{
					// hold this client in the list
					OldHubSearchClient = HubSearchClient;
				}
				else
				{
					HubSearchClient->Client->SetCallBackFunction(0);

					if ( HubSearchClient->Client->GetConnectionState() != estNONE )
					{
						if ( HubSearchClient->Client->GetConnectionState() != estDISCONNECTING )
						{
							HubSearchClient->Client->Disconnect(TRUE);
						}
						OldHubSearchClient = HubSearchClient;
					}
					else
					{
						HubSearchClient->Client->Stop();
						delete HubSearchClient->Client;
						HubSearchClient->Client = 0;
						pHubSearchClientList->Remove(HubSearchClient);
						delete HubSearchClient;
						HubSearchClient = OldHubSearchClient;
					}
				}
			}
			else
			{
				OldHubSearchClient = HubSearchClient;
			}
		}
	}

	HubSearchClientListThread.UnLock();
}

/** */
void CHubSearch::RemoveClient( CClient * Client )
{
	HubSearchClientListThread.Lock();

	sHubSearchClient * HubSearchClient;

	if ( pHubSearchClientList != 0 )
	{
		HubSearchClient = 0;

		while ( (HubSearchClient = (sHubSearchClient *)pHubSearchClientList->Next((CObject*)HubSearchClient)) != NULL )
		{
			if ( HubSearchClient->Client == Client )
			{
				HubSearchClient->timeHubTimeout = time(0);
				HubSearchClient->Client->SetCallBackFunction(0);
				HubSearchClient->bRemove = TRUE;
				break;
			}
		}
	}

	HubSearchClientListThread.UnLock();
}

/** */
void CHubSearch::NewClient()
{
	int l;
	CString server,addr,s;
	sHubSearchClient * HubSearchClient;

	if ( pHubServerList == 0 )
	{
		return;
	}

	// done ?
	if ( pHubServerList->Count() == iCurrentHubServer )
	{
		SendDebug("Serverlist done\n");

		ehSearchType = ehstWAITTIMEOUT;

		return;
	}

	HubSearchClientListThread.Lock();

	l = pHubSearchClientList->Count();

	HubSearchClientListThread.UnLock();

	for(;(l<iMaxThreads)&&(iCurrentHubServer<pHubServerList->Count());l++)
	{
		CClient * Client = new CClient();

		HubSearchClient = new sHubSearchClient;

		HubSearchClient->Client = Client;
		HubSearchClient->bRemove = FALSE;
		HubSearchClient->timeHubTimeout = time(0);

		if ( CConfig::Instance()->GetMode() == ecmPassive )
		{
			HubSearchClient->timeHubTimeout += 30;
		}

		Client->SetNick(CConfig::Instance()->GetSearchNick());
		Client->SetComment(CConfig::Instance()->GetDescription(TRUE));
		Client->SetConnectionType(CConfig::Instance()->GetSpeed());
		Client->SetEMail(CConfig::Instance()->GetEMail());
		Client->SetVersion(VERSION);
		Client->SetShareSize(CString().setNum(CConfig::Instance()->GetShareSize()));
		Client->SetMode(CConfig::Instance()->GetMode());
		Client->SetHandleUserList(m_bHandleUserList);
		Client->SetHandleSearch(FALSE);
		Client->SetHandleMyinfo(FALSE);
		Client->SetHandleForceMove(FALSE);
		Client->SetHandleTransfer(FALSE);
		Client->SetCallBackFunction( new CCallback<CHubSearch>( this, &CHubSearch::DC_ClientCallBack ) );

		// start the thread
//		Client->Start();

		if ( pHubServerList->Next((CObject*&)sCurrentServer) != 0 )
		{
			server = *sCurrentServer;

			iCurrentHubServer++;

			SendDebug( "Search on: [" + CString().setNum(iCurrentHubServer) + "] " + server );

			HubSearchClientListThread.Lock();

			pHubSearchClientList->Add((CObject*)HubSearchClient);

			Client->Connect(server);
		}
		else
		{
			// end on error
			iCurrentHubServer = pHubServerList->Count();
		}

		HubSearchClientListThread.UnLock();
	}
}

/** */
CString CHubSearch::GetSearchString( CMessageSearchFile * msg )
{
	CString s = "";

	if ( msg->bLocal == TRUE )
		s += "Hub:";
	s += msg->sSource;
	s += " ";

	if ( msg->bSizeLimit == FALSE )
		s += "F?";
	else
		s += "T?";

	if ( msg->bSizeAtMost == FALSE )
		s += "F?";
	else
		s += "T?";

	s += CString().setNum(msg->lSize) + "?";
	s += CString().setNum(msg->eFileType) + "?";
	s += msg->sString.Replace(' ',"$") + "|";

	// passive searches with multisearch
	if ( msg->bLocal == TRUE )
		s = "$Search " + s;
	else
		s = "$Search " + s + "$MultiSearch " + s;

	return s;
}

/** send the search to the clients */
int CHubSearch::SendSearch( CString s )
{
	int i = 0;
	CString search;
	CList<DCHubObject> list;
	DCHubObject * HubObject;
	CObject *object = 0;

	while ( (object=m_pSearchList->Next(object)) != 0 )
	{
		switch(((CDCMessage*)object)->m_eType)
		{
			case DC_MESSAGE_SEARCH_USER:
			{
				CMessageSearchUser * msg = (CMessageSearchUser*)object;

				if ( CServerManager::Instance()->IsUserOnline( msg->sNick, s, &list ) == TRUE )
				{
					HubObject = 0;

					while ( (HubObject=list.Next(HubObject)) != 0 )
					{
						CMessageSearchResultUser * MessageSearchResultUser = new CMessageSearchResultUser();

						MessageSearchResultUser->m_eType    = DC_MESSAGE_SEARCHRESULT_USER;
						MessageSearchResultUser->m_sHubName = HubObject->m_sHubName;
						MessageSearchResultUser->sNick      = msg->sNick;

						SendSearchResult(MessageSearchResultUser);
					}

					list.Clear();
				}

				i = 1;

				break;
			}

			case DC_MESSAGE_SEARCH_FILE:
			{
				CMessageSearchFile * msg = (CMessageSearchFile*)object;

				search = GetSearchString(msg);

				if ( search != "" )
					i = CServerManager::Instance()->SendStringToConnectedServers(search,s);
				break;
			}

			default:
			{
				break;
			}
		}
	}

	return i;
}

/** send the search to this client */
void CHubSearch::SendSearch( CClient * Client )
{
	CString search;
	CObject *object = 0;

	while ( (object=m_pSearchList->Next(object)) != 0 )
	{
		switch(((CDCMessage*)object)->m_eType)
		{
			case DC_MESSAGE_SEARCH_USER:
			{
				CMessageSearchUser * msg = (CMessageSearchUser*)object;

				if ( Client->IsUserOnline( msg->sNick ) == TRUE )
				{
					CMessageSearchResultUser * MessageSearchResultUser = new CMessageSearchResultUser();

					MessageSearchResultUser->m_eType    = DC_MESSAGE_SEARCHRESULT_USER;
					MessageSearchResultUser->m_sHubName = Client->GetHubName();
					MessageSearchResultUser->sNick      = msg->sNick;

					SendSearchResult(MessageSearchResultUser);
				}

				break;
			}

			case DC_MESSAGE_SEARCH_FILE:
			{
				CMessageSearchFile * msg = (CMessageSearchFile*)object;

				search = GetSearchString(msg);

				if ( search != "" )
					Client->SendString(search);
				break;
			}

			default:
			{
				break;
			}
		}
	}
}

/** */
int CHubSearch::DC_ClientCallBack( CObject * Client, CObject * Object )
{
	_pHubSearch->ClientThread.Lock();

	CClient * c = (CClient*)Client;
	CObject * o = Object;

	if ( (c == 0) || (o == 0) )
	{
		_pHubSearch->ClientThread.UnLock();
		return -1;
	}

	CDCMessage *DCMsg;

	DCMsg = (CDCMessage*) o;

	switch ( DCMsg->m_eType )
	{
		case DC_MESSAGE_CONNECTION_STATE:
		{
			CMessageConnectionState *msg = (CMessageConnectionState*) o;

			switch(msg->eState)
			{
				case estCONNECTED:
				{
					break;
				}

				case estSOCKETERROR:
				{
					_pHubSearch->SendDebug( "SocketError on " + c->GetIP() + CString(":") + CString().setNum(c->GetPort()) + " [" + msg->sMessage + "]");
					break;
				}

				case estDISCONNECTED:
				{
					_pHubSearch->RemoveClient(c);
				}

				default:
				{
					break;
				}
			}

			break;
		}

		case DC_MESSAGE_VALIDATEDENIDE:
		{
			_pHubSearch->SendDebug( "Validate denide " + c->GetIP() + CString(":") + CString().setNum(c->GetPort()) );
			c->Disconnect(TRUE);
			break;
		}

		case DC_MESSAGE_HUBISFULL:
		{
			_pHubSearch->SendDebug( "Hub is full on " + c->GetIP() + CString(":") + CString().setNum(c->GetPort()) );
			c->Disconnect(TRUE);
			break;
		}

		case DC_MESSAGE_FORCEMOVE:
		{
			_pHubSearch->SendDebug( "Force move on " + c->GetIP() + CString(":") + CString().setNum(c->GetPort()) );
			c->Disconnect(TRUE);
			break;
		}

		case DC_MESSAGE_LOCK:
		case DC_MESSAGE_HUBNAME:
		case DC_MESSAGE_CHAT:
		case DC_MESSAGE_MYNICK:
		case DC_MESSAGE_PRIVATECHAT:
		case DC_MESSAGE_SEARCH:
		case DC_MESSAGE_HELLO:
		case DC_MESSAGE_QUIT:
		case DC_MESSAGE_GETINFO:
		{
			break;
		}

		case DC_MESSAGE_MYINFO:
		{
			CString s;
			CMessageMyInfo * MessageMyInfo = (CMessageMyInfo *)o;

			if ( MessageMyInfo->sNick == CConfig::Instance()->GetSearchNick() )
			{
				if ( _pHubSearch->m_bHandleUserList == FALSE )
				{
					_pHubSearch->SendSearch(c);

					// remove the client only for active searches
					if ( CConfig::Instance()->GetMode() != ecmPassive )
						_pHubSearch->RemoveClient(c);
				}
			}
			break;
		}

		case DC_MESSAGE_NICKLIST:
		{
			if ( _pHubSearch->m_bHandleUserList == TRUE )
			{
				_pHubSearch->SendSearch(c);

				// remove the client only for active searches
				if ( CConfig::Instance()->GetMode() != ecmPassive )
					_pHubSearch->RemoveClient(c);
			}
			break;
		}

		case DC_MESSAGE_SEARCHRESULT:
		{
			if ( CConfig::Instance()->GetMode() == ecmPassive )
			{
				_pHubSearch->SendSearchResult(o);
				o = 0;
			}
			break;
		}

		default:
		{
			printf("callback: %d\n",DCMsg->m_eType);
			break;
		}
	}

	if ( o )
		delete o;

	_pHubSearch->ClientThread.UnLock();

	return 0;
}
