/***************************************************************************
                          cservermanager.cpp  -  description
                             -------------------
    begin                : Don Mai 16 2002
    copyright            : (C) 2002-2003 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <dclib/cclient.h>
#include <dclib/cconfig.h>
#include <dclib/cmanager.h>
#include <dclib/cdownloadmanager.h>

#ifndef WIN32
#include <unistd.h>
#include <stdlib.h>
#endif

#include "cservermanager.h"

/** */
CServerManager::CServerManager()
{
	pClientList = new CThreadList<CClient>();

	m_pCallback = new CCallback<CServerManager>( this, &CServerManager::Callback );
	CManager::Instance()->Add( m_pCallback );

	SetInstance(this);
}

/** */
CServerManager::~CServerManager()
{
	CManager::Instance()->Remove( m_pCallback );
	if (m_pCallback)
	{
		delete m_pCallback;
		m_pCallback = 0;
	}
	
	m_Thread.Lock();
	pClientList->Lock();

	delete pClientList;
	pClientList = 0;

	m_Thread.UnLock();
}

/** thread callbackfunction */
int CServerManager::Callback( CObject *, CObject * )
{
	// thread ...
	CClient * client = 0;

	m_Thread.Lock();

	if ( pClientList )
	{
		while( (client=pClientList->Next(client)) != 0 )
			client->Thread(0);
	}

	m_Thread.UnLock();

	return 0;
}

/** */
void CServerManager::Connect( CString hubname, CString server, CClient * client )
{
	if ( !client )
	{
		if ( server == "" )
		{
			return;
		}

		if ( hubname == "" )
		{
			hubname = server;
		}


		m_Thread.Lock();

		client = (CClient*)GetHub(hubname);

		m_Thread.UnLock();

		if ( client )
		{
			return;
		}

		client = new CClient();
	}

	client->SetHubName(hubname);

	AddHub(client);

	client->SetNick(CConfig::Instance()->GetNick());
	client->SetComment(CConfig::Instance()->GetDescription());
	client->SetConnectionType(CConfig::Instance()->GetSpeed());
	client->SetEMail(CConfig::Instance()->GetEMail());
	client->SetVersion(CString("DCGUI-") + VERSION);
	client->SetShareSize(CString().setNum(CConfig::Instance()->GetShareSize()));
	client->SetMode(CConfig::Instance()->GetMode());
	client->SetAwayMode(CConfig::Instance()->GetAwayMode());
	client->Connect(server);
}

/** */
void CServerManager::AddHub( CClient * client )
{
	m_Thread.Lock();

	if ( pClientList )
	{
		pClientList->Lock();
		pClientList->Add(client);
		pClientList->UnLock();
	}

	m_Thread.UnLock();
}

/** */
void CServerManager::RemoveHub( CClient * client )
{
	m_Thread.Lock();

	if ( pClientList )
	{
		pClientList->Lock();
		client->SetCallBackFunction(0);
		pClientList->Remove(client);
		pClientList->UnLock();
	}

	m_Thread.UnLock();
}

/** */
int CServerManager::SM_ClientCallBack( CObject * /*Sender*/, CObject * /*Object*/ )
{
//	Transfer->SetCallBackFunction( new CCallback<CDownloadManager>( pDownloadManager, &CDownloadManager::DM_TransferCallBack ) );
	return -1;
}

/** */
long CServerManager::GetConnectedHubCount( bool admin )
{
	long i;
	CClient * client;

	if ( !pClientList )
	{
		return 0;
	}

	pClientList->Lock();

	i = 0;
	client = 0;

	while( (client=pClientList->Next(client)) != 0 )
	{
		if ( (client->IsHandshake() == TRUE) )
			continue;

		if ( (admin == TRUE) && (client->IsAdmin() == TRUE) )
			continue;

		i++;
	}

	pClientList->UnLock();

	return i;
}

/** */
CClient * CServerManager::GetHub( CString hubname )
{
	CClient* client = 0;

	if ( !pClientList )
	{
		return client;
	}

	pClientList->Lock();

	while( (client=pClientList->Next(client)) != 0 )
	{
		if ( client->GetHubName() == hubname )
		{
			break;
		}
	}

	pClientList->UnLock();

	return client;
}

/** */
CStringList * CServerManager::GetPublicHubServerList()
{
	DCConfigHubItem * hubitem;
	CStringList * StringList = 0;
	CObject * object;

	CList<DCConfigHubItem> List;

	CConfig::Instance()->GetPublicHubList(&List);

	if ( List.Count() > 0 )
	{
		hubitem = 0;
		StringList = new CStringList();

		while( (hubitem=List.Next(hubitem)) != 0 )
		{
			if ( StringList->Get( hubitem->sHost, (CObject *&)object ) != 0 )
			{
				StringList->Add(hubitem->sHost, new CString(hubitem->sHost));
			}
		}
	}

	return StringList;
}

/** */
CStringList * CServerManager::GetBookmarkHubServerList()
{
	DCConfigHubItem * hubitem;
	CStringList * StringList = 0;
	CObject * object;

	CList<DCConfigHubItem> List;

	CConfig::Instance()->GetBookmarkHubList(&List);

	if ( List.Count() > 0 )
	{
		hubitem = 0;
		StringList = new CStringList();

		while( (hubitem=List.Next(hubitem)) != 0 )
		{
			if ( StringList->Get( hubitem->sHost, (CObject *&)object ) != 0 )
			{
				StringList->Add(hubitem->sHost, new CString(hubitem->sHost));
			}
		}
	}

	return StringList;
}

/** */
CStringList * CServerManager::GetConnectedHubServerList()
{
	CClient* client;
	CStringList * StringList;

	if ( !pClientList )
	{
		return 0;
	}

	pClientList->Lock();

 	StringList = new CStringList();

	if ( pClientList->Count() > 0 )
	{
		client = 0;
		while( (client=pClientList->Next(client)) != 0 )
		{
			if ( client->IsHandshake() == TRUE )
			{
				continue;
			}

			StringList->Add("",new CString(client->GetHubName()));
		}
	}

	pClientList->UnLock();

	return StringList;
}

/** send message to all connected servers */
int CServerManager::SendStringToConnectedServers( CString s, CString hubname )
{
	int i;
	CClient* client;

	if ( !pClientList )
	{
		return 0;
	}

	pClientList->Lock();

	i = 0;

	if ( s != "" )
	{
		client = 0;

		while( (client=pClientList->Next(client)) != 0 )
		{
			if ( client->IsHandshake() == TRUE )
			{
				continue;
			}

			if ( hubname != "" )
			{
				if ( hubname == client->GetHubName() )
				{
					client->SendString(s);
					i++;
					break;
				}
				else
				{
					continue;
				}
			}
			else
			{
				client->SendString(s);
			}

			i++;
		}
	}

	pClientList->UnLock();

	return i;
}

/** send myinfo to all connected servers */
int CServerManager::SendMyInfoToConnectedServers()
{
	CClient* client;
	CString nick,description,speed,email,share;
	eUserAwayMode awaymode;
	int i;

	if ( !pClientList )
	{
		return 0;
	}

	nick        = CConfig::Instance()->GetNick();
	description = CConfig::Instance()->GetDescription();
	speed       = CConfig::Instance()->GetSpeed();
	email       = CConfig::Instance()->GetEMail();
	share       = CString().setNum(CConfig::Instance()->GetShareSize());
	awaymode    = CConfig::Instance()->GetAwayMode();

	pClientList->Lock();

	i = 0;
	client = 0;

	while( (client=pClientList->Next(client)) != 0 )
	{
		if ( client->IsHandshake() == TRUE )
		{
			continue;
		}

		// update internal values
		client->SetNick(nick);
		client->SetComment(description);
		client->SetConnectionType(speed);
		client->SetEMail(email);
		client->SetShareSize(share);
		client->SetAwayMode(awaymode);
		i++;
	}
	pClientList->UnLock();

	return i;
}

/** */
int CServerManager::SendConnectionRequest( CString nick, CString hubname )
{
	int res;
	CClient* client;

	if ( !pClientList )
	{
		return 0;
	}

	pClientList->Lock();

	res = -3;
	client = 0;

	while( (client=pClientList->Next(client)) != 0 )
	{
		if ( client->GetHubName() == hubname )
		{
			if ( client->IsHandshake() == TRUE )
			{
				res = -2;
				break;
			}

			if ( client->IsUserOnline( nick ) == FALSE )
			{
				res = -1;
				break;
			}

			if ( client->GetMode() == ecmPassive )
			{
				// send connection request
				if ( client->SendRevConnectToMe( client->GetNick(), nick ) == 0 )
				{
					res = 0;
				}
				else
				{
					res = -4;
				}
			}
			else if ( client->GetMode() == ecmActive )
			{
				CString s = CConfig::Instance()->GetTCPHostString();

				if ( s != "" )
				{
					// send connection request
					CDownloadManager::Instance()->DLM_AddTransferRequest( nick, "", hubname, client->GetHost() );

					if ( client->SendConnectToMe( nick, s ) == 0 )
					{
						res = 0;
					}
					else
					{
						res = -4;
					}
				}
			}

			break;
		}
	}
	pClientList->UnLock();

	return res;
}

/** */
bool CServerManager::IsUserOnline( CString nick, CString hubname, CList<DCHubObject> * list )
{
	bool res;
	CClient* client;
	DCHubObject * HubObject;

	if ( !pClientList )
	{
		return FALSE;
	}

	pClientList->Lock();

	res = FALSE;
	client = 0;

	while( (client=pClientList->Next(client)) != 0 )
	{
		if ( client->IsHandshake() == FALSE )
		{
			if ( hubname == "" )
			{
				if ( client->IsUserOnline( nick ) == TRUE )
				{
					if ( list )
					{
						HubObject = new DCHubObject();
						HubObject->m_sHubName = client->GetHubName();
						HubObject->m_sHubHost = client->GetHost();
						list->Add(HubObject);
					}
					res = TRUE;
				}
			}
			else if ( client->GetHubName() == hubname )
			{
				if ( client->IsUserOnline( nick ) == TRUE )
					res = TRUE;
				break;
			}
		}
	}
	pClientList->UnLock();

	return res;
}

/** */
eHubState CServerManager::IsHubOnline( CString hubname )
{
	eHubState res;
	CClient* client;

	if ( !pClientList )
	{
		return ehsNONE;
	}

	pClientList->Lock();

	res = ehsNONE;
	client = 0;

	while( (client=pClientList->Next(client)) != 0 )
	{
		if ( client->GetHubName() == hubname )
		{
			if ( client->IsHandshake() == FALSE )
			{
				res = ehsONLINE;
			}
			else
			{
				res = ehsOFFLINE;
			}

			break;
		}
	}
	pClientList->UnLock();

	return res;
}

/** */
bool CServerManager::IsAdmin( CString hubname, CString nick )
{
	bool res;
	CClient* client;

	if ( !pClientList )
	{
		return FALSE;
	}

	pClientList->Lock();

	res = FALSE;
	client = 0;

	while( (client=pClientList->Next(client)) != 0 )
	{
		if ( client->GetHubName() == hubname )
		{
			if ( client->IsHandshake() == FALSE )
			{
				if ( nick == "" )
					res = client->IsAdmin();
				else
					res = client->IsAdmin(nick);
			}

			break;
		}
	}
	pClientList->UnLock();

	return res;
}

/** */
CString CServerManager::GetHubHost( CString hubname )
{
	CString res;
	CClient* client;

	if ( !pClientList )
	{
		return "";
	}

	pClientList->Lock();

 	res = "";
	client = 0;

	while( (client=pClientList->Next(client)) != 0 )
	{
		if ( client->GetHubName() == hubname )
		{
			res = client->GetIP() + ":" + CString().setNum(client->GetPort());
			break;
		}
	}
	pClientList->UnLock();

	return res;
}
