/***************************************************************************
                          csocket.cpp  -  description
                             -------------------
    begin                : Sat Oct 6 2001
    copyright            : (C) 2001-2003 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

#ifndef WIN32
#include <netdb.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>

// includes for sun os
#ifdef HAVE_SYS_SOCKIO_H 
#include <sys/sockio.h>
#endif
#ifdef HAVE_SYS_TERMIOS_H
#include <sys/termios.h>
#endif
#ifdef HAVE_SYS_FILIO_H 
#include <sys/filio.h>
#endif
#endif

#include <sys/types.h>

#ifdef WIN32
#include <winsock2.h>
#endif

#ifndef MSG_NOSIGNAL
// 0x2000  /* don't raise SIGPIPE */
#define MSG_NOSIGNAL      0
#endif

#include <dclib/dcos.h>
#include <dclib/casyncdns.h>
#include <dclib/clogfile.h>
#include <dclib/cssl.h>

#include "csocket.h"

#ifdef HAVE_SOCKS
extern "C" {
#include <socks.h>
}
#endif

CTraffic CSocket::m_Traffic;

/** static critical section*/
static CThread m_cs_ghbn;

CSocket::CSocket( eSocketType type )
{
	SocketType    = type;
	iHandle       = INVALID_SOCKET;
	sError        = "";
	m_eSocketMode = esmSOCKET;
	m_bSSLInit    = FALSE;
#ifdef HAVE_SSL
	m_pCTX = 0;
	m_pSSL = 0;
#endif

#ifdef WIN32
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;
 
	wVersionRequested = MAKEWORD( 2, 2 );
 
	err = WSAStartup( wVersionRequested, &wsaData );

	if ( err != 0 )
	{
		/* Tell the user that we could not find a usable */
		/* WinSock DLL.                                  */
		return;
	}
 
	/* Confirm that the WinSock DLL supports 2.2.*/
	/* Note that if the DLL supports versions greater    */
	/* than 2.2 in addition to 2.2, it will still return */
	/* 2.2 in wVersion since that is the version we      */
	/* requested.                                        */
 
	if ( LOBYTE( wsaData.wVersion ) != 2 ||
		HIBYTE( wsaData.wVersion ) != 2 )
	{
		/* Tell the user that we could not find a usable */
		/* WinSock DLL.                                  */
		WSACleanup( );
		return; 
	}
 
	/* The WinSock DLL is acceptable. Proceed. */
#endif
}

CSocket::~CSocket()
{
	if ( iHandle != INVALID_SOCKET)
	{
#ifdef WIN32
		closesocket(iHandle);
#else
		close(iHandle);
#endif
	}
}

/** */
int CSocket::SetSocket( int handle, eSocketType sockettype )
{
	if ( handle == -1 )
	{
		return -1;
	}

	if ( (sockettype != estTCP) && (sockettype != estUDP) )
	{
		return -1;
	}

	SocketType = sockettype;
	iHandle    = handle;

	return 0;
}

/** rfc1918 */
bool CSocket::IsPrivateAddressSpace( const char * cp )
{
	unsigned long int i = inet_addr(cp);

	if ( i == (unsigned long int)-1 )
		return FALSE;

	/** 10.0.0.0 - 10.255.255.255
	  * 172.16.0.0 - 172.31.255.255
	  * 192.168.0.0 - 192.168.255.255
	*/
	if ( ((i | inet_addr("10.255.255.255"))  == inet_addr("10.255.255.255")) ||
	     ((i | inet_addr("172.16.255.255"))  == inet_addr("172.31.255.255")) ||
	     ((i | inet_addr("192.168.255.255")) == inet_addr("192.168.255.255")) )
		return TRUE;

	return FALSE;
}

/** */
bool CSocket::GetHostByName( const char * name, struct hostent *& hp )
{
	m_cs_ghbn.Lock();

	bool err = FALSE;

	if ( (hp = gethostbyname(name)) != 0 )
		err = TRUE;

	m_cs_ghbn.UnLock();

	return err;
}

/** */
CString CSocket::GetHostByName( CString Host )
{
	CString s = "";
	struct hostent * hp;
	struct sockaddr_in sin;

	if ( Host != "" )
	{
		if ( GetHostByName( Host.Data(), hp ) == TRUE )
		{
			if ( hp->h_addr != 0 )
			{
			 	memcpy((void *)&sin.sin_addr, hp->h_addr, sizeof(sin.sin_addr) ); //hp->h_length);

				s = inet_ntoa(sin.sin_addr);
			}
		}
	}

	return s;
}

/** */
int CSocket::GetInterfaceList( CList<CString> * interfacelist )
{
	if ( !interfacelist )
	{
		return -1;
	}

#ifdef WIN32
	// get the interfaces from win32
	LONG Status;
	ULONG Result;
	HKEY AdapKey,OneAdapKey;
	TCHAR AdapName[256];
	DWORD dim;
	UCHAR buffer[256];
	CString name,instance;
	int i;

	Status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
		TEXT("SYSTEM\\CurrentControlSet\\Control\\Class\\{4D36E972-E325-11CE-BFC1-08002BE10318}"),
		0,
		KEY_READ,
		&AdapKey);

	if ( Status == ERROR_SUCCESS )
	{
		i=0;

		// Get the size to allocate for the original device names
		while( (Result=RegEnumKey(AdapKey,i,AdapName,sizeof(AdapName)/2)) == ERROR_SUCCESS )
		{
			Status = RegOpenKeyEx(AdapKey,AdapName,0,KEY_READ,&OneAdapKey);

			if ( Status == ERROR_SUCCESS )
			{
				name     = "";
				instance = "";

				dim  = 256;
				Status = RegQueryValueEx(OneAdapKey,"NetCfgInstanceId",NULL,NULL,buffer,&dim);

				if ( Status == ERROR_SUCCESS )
				{
					dim = 256;
					instance = (char*)buffer;

					Status = RegQueryValueEx(OneAdapKey,"DriverDesc",NULL,NULL,buffer,&dim);

					if ( Status == ERROR_SUCCESS )
					{
						name = (char*)buffer;
					}
				}

				RegCloseKey(OneAdapKey);

				if ( (instance != "") && (name != "") )
				{
					// now we check if the interface exist
					instance  = "SYSTEM\\CurrentControlSet\\Services\\"+instance +"\\Parameters\\Tcpip";

					Status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
							TEXT(instance.Data()),
							0,
							KEY_READ,
							&OneAdapKey);

					if ( Status == ERROR_SUCCESS )
					{
						interfacelist->Add( new CString(name) );
						RegCloseKey(OneAdapKey);
					}
				}
			}

			i++;
		}

		RegCloseKey(AdapKey);
	}
#else
	int sock = -1;
	struct ifreq ifreqa[256];
	int i;
	int count;
	struct ifconf *pifconf;
	struct ifreq * pifreq;

	sock = socket( PF_INET, SOCK_DGRAM, 0 );

	if ( sock != -1 )
	{
		pifconf = (struct ifconf*) malloc(sizeof(ifconf));

		pifconf->ifc_len = 255*32;
		pifconf->ifc_ifcu.ifcu_req = ifreqa;

		i = ioctl( sock, SIOCGIFCONF, pifconf );

		if ( i != -1 )
		{
			if ( pifconf->ifc_len > 0 )
			{
				count = pifconf->ifc_len / sizeof( struct ifreq );

				pifreq = pifconf->ifc_ifcu.ifcu_req;

				for(i=0;i<count;i++)
				{
					interfacelist->Add( new CString((char*)pifreq[i].ifr_name) );
				}
			}
		}

		close(sock);
		free(pifconf);
	}
#endif

	return interfacelist->Count();
}

/** */
CString CSocket::GetInterfaceIP( CString iface )
{
	CString s = "";
#ifdef WIN32
	// get the interfaces from win32
	LONG Status;
	ULONG Result;
	HKEY AdapKey,OneAdapKey;
	TCHAR AdapName[256];
	DWORD dim;
	DWORD RegType;
	DWORD DHCPEnabled; 
	UCHAR buffer[256];
	CString name,instance;
	int i;

	Status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
		TEXT("SYSTEM\\CurrentControlSet\\Control\\Class\\{4D36E972-E325-11CE-BFC1-08002BE10318}"),
		0,
		KEY_READ,
		&AdapKey);

	if ( Status == ERROR_SUCCESS )
	{
		i=0;

		// Get the size to allocate for the original device names
		while( (Result=RegEnumKey(AdapKey,i,AdapName,sizeof(AdapName)/2)) == ERROR_SUCCESS )
		{
			Status = RegOpenKeyEx(AdapKey,AdapName,0,KEY_READ,&OneAdapKey);

			if ( Status == ERROR_SUCCESS )
			{
				name     = "";
				instance = "";

				dim = 256;
				Status = RegQueryValueEx(OneAdapKey,"DriverDesc",NULL,NULL,buffer,&dim);

				if ( Status == ERROR_SUCCESS )
				{
					name = (char*)buffer;

					if ( name == iface )
					{
						// found
						dim = 256;
						Status = RegQueryValueEx(OneAdapKey,"NetCfgInstanceId",NULL,NULL,buffer,&dim);					

						if ( Status == ERROR_SUCCESS )
						{
							instance = (char*)buffer;
						}
					}
				}

				RegCloseKey(OneAdapKey);

				if ( (instance != "") && (name != "") )
				{
					// now we get the ip
					instance = "SYSTEM\\CurrentControlSet\\Services\\"+instance+"\\Parameters\\Tcpip";

					Status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
							TEXT(instance.Data()),
							0,
							KEY_READ,
							&OneAdapKey);

					if ( Status == ERROR_SUCCESS )
					{
						dim = 4;
						Status = RegQueryValueEx(OneAdapKey,TEXT("EnableDHCP"),NULL,&RegType,(LPBYTE)&DHCPEnabled,&dim);

						if ( Status != ERROR_SUCCESS )
						{
							DHCPEnabled = 0;
						}

						if ( DHCPEnabled == 0 )
						{
							dim = 265;
							Status=RegQueryValueEx(OneAdapKey,"IPAddress",NULL,NULL,buffer,&dim);
							if ( Status == ERROR_SUCCESS )
							{
								s = (char*)buffer;
							}
						}
						else
						{
							dim = 256;
							Status = RegQueryValueEx(OneAdapKey,"DhcpIPAddress",NULL,NULL,buffer,&dim);

							if ( Status == ERROR_SUCCESS )
							{
								s = (char*)buffer;
							}
						}

						RegCloseKey(OneAdapKey);

						break;
					}
				}
			}

			i++;
		}

		RegCloseKey(AdapKey);
	}
#else
	int sock = -1;
	struct ifreq ifreqa[256];
	int i;
	int count;
	struct ifconf *pifconf;
	struct ifreq * pifreq;
	struct ifreq ifr;

	sock = socket( PF_INET, SOCK_DGRAM, 0 );

	if ( sock == -1 )
	{
		return s;
	}

	pifconf = (struct ifconf*) malloc(sizeof(ifconf));

	pifconf->ifc_len = 255*32;
	pifconf->ifc_ifcu.ifcu_req = ifreqa;

	i = ioctl( sock, SIOCGIFCONF, pifconf );

	if ( i == -1 )
	{
		close(sock);
		free(pifconf);
		return s;
	}

	if ( pifconf->ifc_len == 0 )
	{
		close(sock);
		free(pifconf);
		return s;
	}

	count = pifconf->ifc_len / sizeof( struct ifreq );

	pifreq = pifconf->ifc_ifcu.ifcu_req;

	for(i=0;i<count;i++)
	{
// change for freebsd ...
//		if ( iface == (char*)pifreq[i].ifr_ifrn.ifrn_name )
//		{
//			memcpy( ifr.ifr_name,pifreq[i].ifr_ifrn.ifrn_name,16);
//
		if ( iface == (char*)pifreq[i].ifr_name )
		{
			memcpy (ifr.ifr_name,pifreq[i].ifr_name,16);

			if ( ioctl( sock, SIOCGIFADDR, &ifr ) != -1 )
			{
				s = inet_ntoa(((sockaddr_in*)&ifr.ifr_broadaddr)->sin_addr);
			}

			break;
		}
	}

	close(sock);
	free(pifconf);
#endif
	return s;
}

/** */
int CSocket::GetFreeSendBufferSize()
{
	int tot, free;
#ifndef WIN32
	int unsent;
#endif
	socklen_t ilen;

	free = 0;

	ilen = sizeof(int);

	if ( getsockopt(iHandle, SOL_SOCKET, SO_SNDBUF, (char*)&tot, &ilen) == 0 )
	{
		// quick & dirty fix for free bsd
		free = tot;
#ifndef WIN32
		if ( ioctl(iHandle, TIOCOUTQ, &unsent) == 0 )
		{
			free = tot - unsent;
		}
#endif
	}

	return free;
}

/** */
int CSocket::IsConnect()
{
	int i,err;
	struct timeval tv;
	fd_set readset;

	if ( iHandle == INVALID_SOCKET )
	{
		return -1;
	}

	FD_ZERO(&readset);
	FD_SET(iHandle, &readset);

	tv.tv_sec  = 0;
	tv.tv_usec = 1;

	err = -1;

	i = select(FD_SETSIZE, &readset, &readset, NULL, &tv);

	FD_CLR(iHandle, &readset);

	if ( i > 0 )
	{
		err = 1;
	}
	else if ( i == 0 )
	{
		err = 0;
	}

#ifdef HAVE_SSL
	if ( (err != -1) && (m_eSocketMode != esmSOCKET) && (m_bSSLInit == FALSE))
	{
		if ( m_eSocketMode == esmSSLCLIENT )
		{
			err = SSL_connect(m_pSSL);

			if ( err != 1 )
			{
				ERR_print_errors_fp(stderr);
				err = SSL_get_error(m_pSSL,err);

				if ( (err != SSL_ERROR_WANT_READ) && (err != SSL_ERROR_WANT_WRITE) )
				{
					err = -1;
					ERR_print_errors_fp(stderr);
				}
				else
				{
					err = 0;
				}
			}
			else
			{
				printf("Connected with %s encryption\n", SSL_get_cipher(m_pSSL));
				SSL_get_peer_certificate(m_pSSL);
				m_bSSLInit = TRUE;
			}
		}
		else if ( m_eSocketMode == esmSSLSERVER )
		{
			err = SSL_accept(m_pSSL);

			if ( err != 1 )
			{
				ERR_print_errors_fp(stderr);
				err = SSL_get_error(m_pSSL,err);

				if ( (err != SSL_ERROR_WANT_READ) && (err != SSL_ERROR_WANT_WRITE) )
				{
					err = -1;
					ERR_print_errors_fp(stderr);
				}
				else
				{
					err = 0;
				}
			}
			else
			{
				printf("Connected with %s encryption\n", SSL_get_cipher(m_pSSL));
				SSL_get_peer_certificate(m_pSSL);
				m_bSSLInit = TRUE;
			}
		}
	}
#endif

	return err;
}

/** */
eConnectState CSocket::Connect( CString Host, bool bAsync )
{
	int i,port;
	CString s = "";

	i = Host.Find(':');

	if ( i > 0 )
	{
		s = Host.Mid(i+1,Host.Length()-i-1);
	}

	if ( (i < 0) || (s == "") )
	{
		port = 411;
		s    = Host;
	}
	else
	{
		port = atoi(s.Data());
		s    = s.Mid(0,i);
	}

	return Connect( s, port, bAsync );
}

/** */
eConnectState CSocket::Connect( CString Host, int Port, bool bAsync )
{
	struct hostent *hp;
	struct sockaddr_in sin;
	SOCKET sockfd;
	eAsyncDns ead;

	if ( iHandle != INVALID_SOCKET )
	{
		Disconnect();
	}

	memset((void *)&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;

	if ( Host == "" )
	{
		if ( SocketType == estTCP )
		{
			return ecsERROR;
		}
		else
		{
			sin.sin_addr.s_addr = htonl(INADDR_ANY);
		}
	}
	else
	{
		if ( bAsync == FALSE )
		{
			// use blocked gethostbyname
			if ( GetHostByName( Host.Data(), hp ) == FALSE )
			{
#ifdef WIN32
				sError = strerror(WSAGetLastError());
#else
				sError = hstrerror(h_errno);
#endif
				return ecsERROR;
			}
			else if ( hp->h_addr == 0 )
			{
				return ecsERROR;
			}
			else
			{
		 		memcpy((void *)&sin.sin_addr, hp->h_addr, hp->h_length);
			}
		}
		else
		{
			// use async dns class
			if ( CAsyncDns::Instance() )
			{
				ead = CAsyncDns::Instance()->GetHostByName( Host, &sin.sin_addr, &sError );
			}
			else
			{
				return ecsERROR;
			}

			if ( ead == eadAGAIN )
			{
				return ecsAGAIN;
			}
			else if ( ead == eadERROR )
			{
				return ecsERROR;
			}
		}
	}

	sin.sin_port = htons((unsigned short)Port);

	// socket
	if ( SocketType == estTCP )
	{
		if ((sockfd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
		{
			sError = strerror(errno);
			return ecsERROR;
		}
	}
	else
	{
		if( (sockfd = socket( PF_INET, SOCK_DGRAM, 0 )) < 0 )
		{
			sError = strerror(errno);
			return ecsERROR;
		}
	}

	// set async flag
#ifdef WIN32
	unsigned long flag = bAsync;
	if ( ioctlsocket(sockfd, FIONBIO, &flag ) != 0 )
#else
	int flag = bAsync;
	if ( ioctl(sockfd, FIONBIO, &flag ) != 0 )
#endif
	{
		sError = strerror(errno);
		return ecsERROR;
	}

	if ( Host != "" )
	{
		// connect
		if ( connect(sockfd, (struct sockaddr *)&sin, sizeof(sin)) != 0 )
		{
#ifdef WIN32
			int e = WSAGetLastError();
			if ( (e != WSAEINPROGRESS) && (e != WSAEWOULDBLOCK) )
#else
			if ( errno != EINPROGRESS )
#endif
			{
				sError = strerror(errno);
#ifdef WIN32
				closesocket(sockfd);
#else
				close(sockfd);
#endif
				return ecsERROR;
			}
		}
	}
	else
	{
		if ( bind(sockfd, (struct sockaddr *) &sin, sizeof(sin)) < 0 )
		{
			sError = strerror(errno);
#ifdef WIN32
			closesocket(sockfd);
#else
			close(sockfd);
#endif
			return ecsERROR;
		}
	}

	iHandle = sockfd;

	return ecsSUCCESS;
}

/** */
int CSocket::Disconnect()
{
	if ( iHandle != INVALID_SOCKET )
	{
#ifdef HAVE_SSL
		if ( (m_eSocketMode != esmSOCKET) && (m_pSSL) )
		{
			SSL_free(m_pSSL);
			m_pSSL = 0;
		}
#endif

#ifdef WIN32
		closesocket(iHandle);
#else
		close(iHandle);
#endif
#ifdef HAVE_SSL
		if ( (m_eSocketMode != esmSOCKET) && (m_pCTX) )
		{
			SSL_CTX_free(m_pCTX);
			m_pCTX = 0;
		}
#endif
		m_eSocketMode = esmSOCKET;
		iHandle       = INVALID_SOCKET;
		m_bSSLInit    = FALSE;
	}

	return 0;
}

/** */
int CSocket::Read( char * buffer, int len, int sec_timeout, int usec_timeout )
{
	int l,i;
	struct timeval tv;
	fd_set readset;
	socklen_t cli_len;
	struct sockaddr cli_addr;

	cli_len = sizeof(cli_addr);

	l = 0;
	i = 0;

	if ( iHandle == INVALID_SOCKET )
	{
		return -1;
	}
#ifdef HAVE_SSL
	if ( (m_eSocketMode == esmSSLCLIENT) || (m_eSocketMode == esmSSLSERVER) )
	{
		l = SSL_read(m_pSSL, buffer, len);

		if ( (l <= 0) && ((i=IsConnect()) != -1) )
		{
//			printf("READ: %d %d\n",l,i);

			l = SSL_get_error(m_pSSL,l);

			if ( (l != SSL_ERROR_WANT_READ) && (l != SSL_ERROR_WANT_WRITE) )
			{
				l = -1;
				ERR_print_errors_fp(stderr);
			}
			else
			{
				l = i = 0;
			}
		}
	}
	else
#endif
	{
		FD_ZERO(&readset);
		FD_SET(iHandle, &readset);

		tv.tv_sec  = sec_timeout;
		tv.tv_usec = usec_timeout;

		i = select(FD_SETSIZE, &readset, NULL, NULL, &tv);

		FD_CLR(iHandle, &readset);

		if ( i > 0 )
		{
			// handle udp socket
			if ( SocketType == estUDP )
			{
				if ( (l=recvfrom(iHandle,buffer,len,0,&cli_addr,&cli_len)) < 0 )
				{
#ifdef WIN32
					int e = WSAGetLastError();
					if ( (e != WSAEINPROGRESS) && (e != WSAEWOULDBLOCK) )
					{
#else
					if ( (errno != EINPROGRESS) && (errno != EAGAIN) )
					{
						sError = strerror(errno);
#endif
					}
					else
					{
						i = l = 0;
					}
				}
			}
			// handle tcp socket
			else if ( SocketType == estTCP )
			{
				if ( (l=recv(iHandle,buffer,len,0)) < 0 )
				{
					sError = strerror(errno);
				}
			}
		}
		else if ( i < 0 )
		{
			l = -1;
			sError = strerror(errno);
		}
	}

#if 0
	if ( l > 0 )
	{
		CString s;
		s = "RECV:" + CString().setNum(l);
		if( l <= len )
		{
			s+="\n";
			buffer[l]=0;
			s += buffer;
		}
		CLogFile::Write("dcsocket.log",eltINFO,s);
	}
#endif

	/** we are disconnected */
	if ( (i==1) && (l==0) )
		l = -1;

	if ( l > 0 )
		CSocket::m_Traffic.AddTraffic( ettRX, l );

	return l;
}

/** */
int CSocket::Write( const unsigned char * buffer, int len, int sec_timeout, int usec_timeout )
{
	int i;
	struct timeval tv;
	fd_set writeset;

	if ( iHandle == INVALID_SOCKET )
	{
		return -1;
	}

#ifdef HAVE_SSL
	if ( (m_eSocketMode == esmSSLCLIENT) || (m_eSocketMode == esmSSLSERVER) )
	{
		i = SSL_write(m_pSSL, buffer, len);

		if ( (i == -1) && (IsConnect() != -1) )
		{
//			printf("WRITE: %d %d\n",i,IsConnect());

			i = SSL_get_error(m_pSSL,i);

			if ( (i != SSL_ERROR_WANT_READ) && (i != SSL_ERROR_WANT_WRITE) )
			{
				i = -1;
				ERR_print_errors_fp(stderr);
			}
			else
			{
				i = 0;
			}
		}
	}
	else
#endif
	{
		FD_ZERO(&writeset);
		FD_SET(iHandle, &writeset);

		tv.tv_sec  = sec_timeout;
		tv.tv_usec = usec_timeout;

		i = select(FD_SETSIZE, NULL, &writeset, NULL, &tv);

		FD_CLR(iHandle, &writeset);

		if ( i > 0 )
		{
			// handle tcp socket
			if ( SocketType == estTCP )
			{
				i = send( iHandle, (const char*)buffer, len, MSG_NOSIGNAL );
			}
			else
			{
				i = send( iHandle, (const char*)buffer, len, MSG_NOSIGNAL );
			}

			// send error or wrong socket type
			if ( i < 0 )
			{
#ifdef WIN32
				int e = WSAGetLastError();
				if ( (e != WSAEINPROGRESS) && (e != WSAEWOULDBLOCK) )
				{
#else
				if ( errno != EINPROGRESS )
				{
					sError = strerror(errno);
#endif
				}
				else
				{
					i = 0;
				}
			}
		}
	}

#if 0
	if ( i > 0 )
	{
		CString s;
		s = "SEND:" + CString().setNum(len)+"\n";
		s += (char*)buffer;
		CLogFile::Write("dcsocket.log",eltINFO,s);
	}
#endif

	if ( i > 0 )
		CSocket::m_Traffic.AddTraffic( ettTX, i );

	return i;
}

/** */
int CSocket::Accept()
{
	struct timeval tv;
	fd_set readset;
	struct sockaddr_in serv_addr;
	SOCKET s = -1;
	int i;
	socklen_t sin_size;

	sin_size = sizeof(struct sockaddr_in);

	FD_ZERO(&readset);
	FD_SET(iHandle, &readset);

	tv.tv_sec  = 0;
	tv.tv_usec = 1;

	i = select(FD_SETSIZE, &readset, NULL, NULL, &tv);

	FD_CLR(iHandle, &readset);

	if ( i > 0 )
	{
		if ((s = accept(iHandle, (struct sockaddr *)&serv_addr, &sin_size)) == INVALID_SOCKET)
		{
			sError = strerror(errno);
			return -1;
		}
		else
		{
			// set async flag
#ifdef WIN32
			unsigned long flag = 1;
			if ( ioctlsocket(s, FIONBIO, &flag ) != 0 )
#else
			int flag = 1;
			if ( ioctl(s, FIONBIO, &flag ) != 0 )
#endif
			{
				sError = strerror(errno);
				return -1;
			}
		}
	}
	
	return s;
}

/** */
int CSocket::Listen( int port )
{
	SOCKET sockfd;
	struct sockaddr_in serv_addr;

	if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET)
	{
#ifdef WIN32
		sError = strerror(WSAGetLastError());
#else
		sError = strerror(errno);
#endif
		return -1;
	}

	/* Let the kernel reuse the socket address. This lets us run
	   twice in a row, without waiting for the (ip, port) tuple
	   to time out. */
#ifdef WIN32
	const char i = 1;
#else
	int i = 1;
#endif
	if ( setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i)) != 0 )
	{
#ifdef WIN32
		sError = strerror(WSAGetLastError());
		closesocket(sockfd);
#else
		sError = strerror(errno);
		close(sockfd);
#endif
		return -1;
	}

	serv_addr.sin_family = AF_INET;
	serv_addr.sin_port = htons((unsigned short)port);
	serv_addr.sin_addr.s_addr = INADDR_ANY;

	if (bind(sockfd, (struct sockaddr *)&serv_addr, sizeof(struct sockaddr)) == -1)
	{
#ifdef WIN32
		sError = strerror(WSAGetLastError());
		closesocket(sockfd);
#else
		sError = strerror(errno);
		close(sockfd);
#endif
		return -1;
	}

	if (listen(sockfd, 5) == -1)
	{
#ifdef WIN32
		sError = strerror(WSAGetLastError());
		closesocket(sockfd);
#else
		sError = strerror(errno);
		close(sockfd);
#endif
		return -1;
	}

	iHandle = sockfd;

	return 0;
}

/** */
bool CSocket::GetPeerName( CString * host, int * port )
{
	struct sockaddr_in addr;
	socklen_t sin_size;

	sin_size = sizeof(struct sockaddr_in);

	if ( (!host) || (!port) )
	{
		return FALSE;
	}

	if ( getpeername( iHandle, (struct sockaddr*)&addr, &sin_size ) == -1 )
	{
		sError = strerror(errno);
		return FALSE;
	}

	*host = inet_ntoa(addr.sin_addr);
	*port = ntohs(addr.sin_port);

	return TRUE;
}

/** */
bool CSocket::ChangeSocketMode( eSocketMode mode, CString cert, CString key )
{
	bool res = FALSE;

	switch(mode)
	{
		case esmSOCKET:
		{
			m_eSocketMode = mode;
			res = TRUE;
			break;
		}
#ifdef HAVE_SSL
		case esmSSLCLIENT:
		case esmSSLSERVER:
		{
			if ( m_eSocketMode == esmSOCKET )
			{
				CSSL cssl;

				if ( mode == esmSSLCLIENT )
				{
					m_pCTX = cssl.InitClientCTX();
				}
				else
				{
					m_pCTX = cssl.InitServerCTX();
					cssl.LoadCertificates(m_pCTX, cert.Data(), key.Data() );
				}

				SSL_CTX_set_mode(m_pCTX, SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER|SSL_MODE_ENABLE_PARTIAL_WRITE);
				m_pSSL = SSL_new(m_pCTX);
				SSL_set_fd(m_pSSL, iHandle);
				m_eSocketMode = mode;
				res = TRUE;
			}

			break;
		}
#endif
		default:
			break;
	}

	return res;
}
